#
# [name] nkj.ncore.py
# [exec] python -m nkj.core
#
# Written by Yoshikazu NAKAJIMA
#

__ROOTPATH = None
_DEBUGLEVEL = 2

import os
import sys
import platform

sys.path.append(os.path.abspath(".."))
import nkj.str as ns
from nkj.str import *

#-- global functions

def rootpath(rootpath=None):
	global __ROOTPATH
	if (rootpath is None):
		if (__ROOTPATH == None):
			__ROOTPATH = os.getenv('NKJ_ROOT')
		return __ROOTPATH
	else:
		__ROOTPATH = rootpath
		return True

def getPythonVersion():
	major = sys.version_info.major
	minor = sys.version_info.minor
	return major, minor

def checkPythonVersion():
	dprint(["--> checkPythonVersion()"])
	vermajor, verminor = getPythonVersion()
	ldprint(["python version: ", str(vermajor), ".", str(verminor)])
	if (vermajor < 3):
		print_error("python version is too old.")
		exit
	elif (vermajor == 3 and verminor < 6):
		print_error("python version is too old.")
		exit
	dprint(["<-- checkPythonVersion()"])

#-- Global initialization

# Append root path

if (rootpath() == None):
	rootpath(os.path.abspath(".."))

sys.path.append(rootpath())

# !Append root path

#-- main

if __name__ == '__main__':
	#ns.debuglevel(_DEBUGLEVEL)
	dprint(["ROOTPATH: ", rootpath()])
	checkPythonVersion()



#
# [name] nkj.cam.py
#
# Written by Yoshikazu NAKAJIMA
#
_LIB_DEBUGLEVEL = 0

import os
import sys
import numpy as np
import json

sys.path.append(os.path.abspath(".."))
from nkj.str import *
from nkj.math import *

ERROR_FOCALLENGTH = -1.0
ERROR_IMAGESIZE = [-1, -1]

_DEFAULT_IMAGECENTER = [0.0, 0.0]
_DEFAULT_IMAGESIZE = ERROR_IMAGESIZE
_DEFAULT_FOCALLENGTH = [ERROR_FOCALLENGTH, ERROR_FOCALLENGTH]
_DEFAULT_CAMERAMATRIX = np.identity(4)
_DEFAULT_FILENAME = None

class camera:
	#-- Class variables
	_classname = "nkj.cam.camera"

	#--
	def __init__(self, focallength=_DEFAULT_FOCALLENGTH):
		ldprint(["--> nkjcam.camera.__init__()"])
		ldprint(["class name: {0}".format(str_bracket(self.getClassName()))])
		self.setImageCenter(_DEFAULT_IMAGECENTER)
		self.setImageSize(_DEFAULT_IMAGESIZE)
		self.setFocalLength(focallength)
		self.setMatrix(_DEFAULT_CAMERAMATRIX)
		self.setFileName(_DEFAULT_FILENAME)
		ldprint(["<-- nkjcam.camera.__init__()"])

	def __str__(self, title=''):
		return self.getPrintStr(title)

	@classmethod
	def getClassName(cls):
		return camera._classname

	@classmethod
	@property
	def classname(cls):
		return cls.getClassName()

	def getPrintStr(self, title=''):
		if (title == ''):
			mesg = '---\n'
		else:
			mesg = '--- {0} ---\n'.format(title)
		filename = self.getFileName()
		if (filename != None):
			mesg += 'File name:    {0}\n'.format(str_bracket(self.getFileName()))
		mesg += 'Focal length: {0}\n'.format(self.getFocalLength())
		mesg += 'Image center: {0}\n'.format(self.getImageCenter())
		mesg += 'Image size:   {0}\n'.format(self.getImageSize())
		mesg += 'Camera matrix:\n'
		mesg += '{0}\n'.format(mat4x4(self.getMatrix()).getPrintStr())
		mesg += '---'
		return mesg

	@property
	def str(self):
		return self.getPrintStr()

	def getImageCenter(self):
		return self._imagecenter

	@property
	def imagecenter(self):
		return self.getImageCenter()

	@property
	def icen(self):
		return self.getImageCenter()

	def getC0(self):
		return self._imagecenter[0]

	@property
	def c0(self):
		return self.getC0()

	def getR0(self):
		return self._imagecenter[1]

	@property
	def r0(self):
		return self.getR0()

	def getImageSize(self):
		return self._imagesize

	@property
	def imagesize(self):
		return self.getImageSize()

	@property
	def isize(self):
		return self.getImageSize()

	def getFocalLength(self):
		return self._focallength

	@property
	def focallength(self):
		return self.getFocalLength()

	@property
	def f(self):
		return self.getFocalLength()

	def getXFocalLength(self):
		return self._focallength[0]

	@property
	def fx(self):
		return self.getXFocalLength()

	def getYFocalLength(self):
		return self._focallength[1]

	@property
	def fy(self):
		return self.getYFocalLength()

	def getAspectRatio(self):
		if (self.focallength[1] == 0):
			___ERROR___
		return self._focallength[0] / self._focallength[1]

	def getAspect(self):
		return self.getAspectRatio()

	@property
	def aspectratio(self):
		return self.getAspectRatio()

	@property
	def aspect(self):
		return self.getAspectRatio()

	def getMatrix(self):
		return self._cameramatrix.getMatrix()

	@property
	def matrix(self):
		return self.getMatrix()

	@property
	def mat(self):
		return self.getMatrix()

	@property
	def m(self):
		return self.getMatrix()

	def getInverseMatrix(self):
		return np.linalg.inv(self.getMatrix())

	def getInvMatrix(self):
		return self.getInverseMatrix()

	def getIMatrix(self):
		return self.getInverseMatrix()

	@property
	def inversematrix(self):
		return self.getInverseMatrix()

	@property
	def invmatrix(self):
		return self.getInverseMatrix()

	@property
	def imatrix(self):
		return self.getInverseMatrix()

	@property
	def invmat(self):
		return self.getInverseMatrix()

	@property
	def imat(self):
		return self.getInverseMatrix()

	@property
	def im(self):
		return self.getInverseMatrix()

	def getFocalPoint(self):
		return self.getMatrix()[0:3, 3]

	@property
	def focalpoint(self):
		return self.getFocalPoint()

	@property
	def fp(self):
		return self.getFocalPoint()

	def getOrigin(self):
		return self.getFocalPoint()

	@property
	def origin(self):
		return self.getOrigin()

	@property
	def orig(self):
		return self.getOrigin()

	def getXaxis(self):
		return self.getMatrix()[0:3, 0]

	@property
	def xaxis(self):
		return self.getXaxis()

	def getYaxis(self):
		return self.getMatrix()[0:3, 1]

	@property
	def yaxis(self):
		return self.getYaxis()

	def getZaxis(self):
		return self.getMatrix()[0:3, 2]

	@property
	def zaxis(self):
		return self.getZaxis()

	def getDirection(self):
		return self.getZaxis()

	@property
	def direction(self):
		return self.getDirection()

	def getLineOfSight(self, p):
		return line3d(self.getOrigin(), p)

	def line_of_sight(self, p):
		return self.getLineOfSight(p)

	def lineofsight(self, p):
		return self.getLineOfSight(p)

	def sight(self, p):
		return self.getLineOfSight(p)

	def getFileName(self):
		return self._filename

	@property
	def filename(self):
		return self.getFileName()

	def setImageCenter(self, imagecenter):
		self._imagecenter = imagecenter

	@imagecenter.setter
	def imagecenter(self, second):
		self.setImageCenter(second)

	@icen.setter
	def icen(self, second):
		self.setImageCenter(second)

	def setImageSize(self, imagesize):
		self._imagesize = imagesize

	@imagesize.setter
	def imagesize(self, second):
		self.setImageSize(second)

	@isize.setter
	def isize(self, second):
		self.setImageSize(second)

	def setFocalLength(self, focallength):
		ldprint2(["--> setFocalLength()"])
		ldprint2(["Focal length: {0}".format(focallength)])
		ldprint2(["Focal length type: {0}".format(type(focallength))])
		if (type(focallength) == list):
			pass
		elif (type(focallength) == float):
			focallength = [focallength, focallength]
		elif (type(focallength) == tuple):
			focallength = list(focallength)
		ldprint2(["> Focal length: {0}".format(focallength)])
		ldprint2(["> Focal length type: {0}".format(type(focallength))])

		if (len(focallength) != 2):
			print_error("Illegal focal length.")
			___ERROR___

		self._focallength = focallength

	@focallength.setter
	def focallength(self, second):
		self.setFocalLength(second)

	@f.setter
	def f(self, second):
		self.setFocalLength(second)

	def setMatrix(self, cm):
		self._cameramatrix = cstrans3d(np.array(cm))

	@matrix.setter
	def matrix(self, second):
		self.setMatrix(second)

	@mat.setter
	def mat(self, second):
		self.setMatrix(second)

	@m.setter
	def m(self, second):
		self.setMatrix(second)

	def setInverseMatrix(self, cm):
		self.setMatrix(np.linalg.inv(cm))

	@inversematrix.setter
	def inversematrix(self, second):
		self.setInverseMatrix(second)

	@invmatrix.setter
	def invmatrix(self, second):
		self.setInverseMatrix(second)

	@imatrix.setter
	def imatrix(self, second):
		self.setInverseMatrix(second)

	@invmat.setter
	def invmat(self, second):
		self.setInverseMatrix(second)

	@imat.setter
	def imat(self, second):
		self.setInverseMatrix(second)

	@im.setter
	def im(self, second):
		self.setInverseMatrix(second)

	def setFileName(self, fname):
		self._filename = fname

	@filename.setter
	def filename(self, second):
		self.setFileName(second)

	def rotateLocal(self, a, flag=None):
		self._cameramatrix.rotateLocal(a, flag)

	def rotate_local(self, a, flag=None):
		self.rotateLocal(a, flag)

	def rotateGlobal(self, a, flag=None):
		self._cameramatrix.rotateGlobal(a, flag)

	def rotate_global(self, a, flag=None):
		self.rotateGlobal(a, flag)

	def translateLocal(self, v):
		self._cameramatrix.translateLocal(v)

	def translate_local(self, v):
		self.translateLocal(v)

	def translateGlobal(self, v):
		self._cameramatrix.translateGlobalLocal(v)

	def translate_glocal(self, v):
		self.translateGlobal(v)

	def projectPoint3D(self, p3d): # 3D点をカメラ画像上に投影した2D点の座標を返す
		#print("p3d: {0}".format(p3d))
		cm = self.getMatrix()
		#print("cm:  {0}".format(cm))
		p3d4 = np.array((p3d[0], p3d[1], p3d[2], 1.0))
		#print("p3d4: {0}".format(p3d4))
		#lp3d = np.linalg.inv(cm) @ p3d4
		lp3d = cm @ p3d4
		#print("lp3d: {0}".format(lp3d))
		return self.projectLocalPoint3D(lp3d)

	def projectLocalPoint3D(self, p3d): # カメラ座標系の3D点をカメラ画像上に投影した2D点の座標を返す
		x3d = p3d[0]
		y3d = p3d[1]
		z3d = p3d[2]
		ldprint(["p3d: ({0}, {1}, {2})".format(x3d, y3d, z3d)])
		f = self.getFocalLength()
		fx = f[0]
		fy = f[1]
		ldprint(["f:   ({0}, {1})".format(fx, fy)])
		ic = self.getImageCenter()
		icx = ic[0]
		icy = ic[1]
		p2d = np.array(((x3d * fx) / z3d + icx, (y3d * fy) / z3d + icy))
		return p2d

	def projectLine3D(self, l3d): # 3D直線をカメラ画像上に投影した2D直線を返す
		l2d = []
		for i in range(2):
			l2d.append(self.projectPoint3D(l3d[i]))
		return l2d

	def load(self, filename, formatindex=None):
		ldprint2(["--> load({0})".format(str_bracket(filename))])
		self.setFileName(filename)
		if (formatindex is None):
			if (os.path.splitext(filename)[1] == '.cam'):
				formatindex = 'sasama'
		with open(filename, 'r') as f:
			if (formatindex == None or formatindex == 'json'):
				data = json.load(f)
				print(data)
				#ndata = {}
				#for key in data.keys():
				#	ndata[key] = ns.extract_float(data[key])
				#ldprint(["<-- load()"])
				#return ndata
				ldprint(["<-- load()"])
				return data
			elif (formatindex == 'yashima.json'):
				data = json.load(f)
				self.setImageSize((data['width'], data['height']))
				intrinsic = data['intrinsic']
				self.setFocalLength((intrinsic[0][0], intrinsic[1][1]))
				self.setImageCenter((intrinsic[0][2], intrinsic[1][2]))
				r = data['rotation']
				t = data['translation']
				self.setMatrix([[r[0][0], r[0][1], r[0][2], t[0]], [r[1][0], r[1][1], r[1][2], t[1]], [r[2][0], r[2][1], r[2][2], t[2]], [0.0, 0.0, 0.0, 1.0]])
				ldprint2(["<-- load()"])
				return data
			elif (formatindex == 'sasama'):
				with open(filename, 'r') as f:
					print('CAUTION: Sasama format was redefined. Check the .cam format when you will load the file made before 2022.03.')
					s_line = f.readlines()
					self.setFocalLength(l2f(s_line[0]))
					self.setImageCenter(l2f(s_line[1]))
					i = 2
					if (len(l2f(s_line[2])) == 2):
						self.setImageSize(l2i(s_line[2]))
						i += 1
					m = []
					for j in range(i, i + 3):
						m.append(l2f(s_line[j]))
						i += 1
					self.setMatrix(m)
					line = l2f(s_line[i])
					if (line == [0.0, 0.0, 0.0, 1.0]):
						i += 1
						line = l2f(s_line[i])
					if (len(s_line) - i > 0):
						#- Non-linear distortion
						print('--> Non-linear distortion')
						print('___NOT_IMPLEMENTED___')
						print('<-- Non-linear distortion')
			else:
				___NOT_SUPPORTED___
		ldprint2(["<-- load()"])

	def print(self, title=''):
		print(self.getPrintStr(title))

def cam2mat(cam):
	if (type(cam) is camera):
		return cam.getMatrix()
	else:
		return cam

# カメラ輻輳角1. カメラY軸（上向きベクトル）周りの3D回転角度．回転でカメラの上向きベクトルがあまり変化しない（＝カメラの上向きベクトル周りに回転している）と仮定．Headding Up/Down が微小であると仮定して，仰俯角をキャンセル．
def convergence_angle(cam1, cam2): # cam1, cam2: カメラクラス or 基準座標系内でのcamera{1, 2}の位置姿勢
	T_cam1 = cam2mat(cam1)
	T_cam2 = cam2mat(cam2)
	cam1_yaxis = T_cam1[0:3, 1] # [0行目から3行分, 1列目(y-axis)]
	cam2_yaxis = T_cam2[0:3, 1]
	cam_yaxis = (cam1_yaxis + cam2_yaxis) / 2.0 # 2 つの y-axis の平均を y-axis とする
	cam1_zaxis = T_cam1[0:3, 2] # [0行目から3行分, 2列目(z-axis)]
	xaxis = np.cross(cam_yaxis, cam1_zaxis)
	cam1_zaxis = np.cross(xaxis, cam_yaxis)
	cam2_zaxis = T_cam2[0:3, 2]
	xaxis = np.cross(cam_yaxis, cam2_zaxis)
	cam2_zaxis = np.cross(xaxis, cam_yaxis)
	return vecangle(cam1_zaxis, cam2_zaxis)

# カメラ輻輳角2. カメラ視線（カメラ中心線）ベクトル間の3D角度．仰俯角（＝Heading Up/Down による輻輳角の変化）も考慮。
def convergence_angle2(cam1, cam2):
	T_cam1 = cam2mat(cam1)
	T_cam2 = cam2mat(cam2)
	cam1_zaxis = T_cam1[0:3, 2]
	cam2_zaxis = T_cam2[0:3, 2]
	return vecangle(cam1_zaxis, cam2_zaxis)

# カメラ輻輳角3. 計画線周りのカメラの3D角度変化．画像に写り込んだ計画線の位置姿勢（画像面内の回転角および画像中心点からの位置ずれ）を考慮．
def convergence_angle3(cam1, cam2):
	T_cam1 = cam2mat(cam1)
	T_cam2 = cam2mat(cam2)
	plan_vec = np.array([0, 0, 1])
	cam1_vec = T_cam1[0:3, 3] # 原点（＝計画線位置）からカメラ焦点へのベクトル
	xaxis = np.cross(plan_vec, cam1_vec)
	cam1_vec = np.cross(xaxis, plan_vec) # 仰角を取り除いたカメラ焦点から計画直線へ向かう（計画直線に垂直な = 最短距離の）ベクトル
	cam2_vec = T_cam2[0:3, 3]
	xaxis = np.cross(plan_vec, cam2_vec)
	cam2_vec = np.cross(xaxis, plan_vec) # 仰角を取り除いたカメラ焦点から計画直線へ向かう（計画直線に垂直な = 最短距離の）ベクトル
	return vecangle(cam1_vec, cam2_vec)

def convergenceangle(T_cam1, T_cam2):
	return convergence_angle(T_cam1, T_cam2)

def convergenceangle2(T_cam1, T_cam2):
	return convergence_angle2(T_cam1, T_cam2)

def convergenceangle3(T_cam1, T_cam2):
	return convergence_angle3(T_cam1, T_cam2)


#-- main

if __name__ == '__main__':
	lib_debuglevel(_LIB_DEBUGLEVEL)

	ldprint(["DEBUG LEVEL: ", lib_debuglevel()])

	print("Class name: \'{0}\'".format(camera.classname))

	cam = camera((1.0, 2.0))

	print("Image center: ", cam.getImageCenter())
	print("Focal length: ", cam.getFocalLength())
	print("Camera matrix:")
	print(cam.getMatrix())

	filename = "./testdata/camera.json"

	if (cam.load(filename, formatindex='yashima.json') is False):
		print_error("Can't load {0}".format(str_bracket(filename)))
		exit(1)

	cam.print('camera parameter')

	p2d = cam.projectPoint3D((0.0, 0.0, 0.0))
	print(p2d)

	cam2 = camera()
	cam2.load('testdata/test.cam')

	cam3 = camera()
	cam3.load('testdata/test2.cam')



#
# [name] nkj.math.py
# [exec] python -m nkj.math
#
# Written by Yoshikazu NAKAJIMA
#
_LIB_DEBUGLEVEL = 0

import os
import sys
import numpy as np
import math
from scipy.spatial.transform import Rotation as R
import copy
import numbers
from functools import singledispatch

sys.path.append(os.path.abspath(".."))
import nkj.str as ns
from nkj.str import *
import nkj.prob as p

#-- flags

ENABLE_MULOPERATOR_FOR_GEOMETRYCOMPUTATION = False
ENABLE_MULOPERATOR_FOR_CSTRANS3D_CLASS = ENABLE_MULOPERATOR_FOR_GEOMETRYCOMPUTATION
ENABLE_MULOPERATOR_FOR_MAT4X4_CLASS = ENABLE_MULOPERATOR_FOR_GEOMETRYCOMPUTATION
ENABLE_MULOPERATOR_FOR_POINT3D_CLASS = ENABLE_MULOPERATOR_FOR_GEOMETRYCOMPUTATION
ENABLE_MULOPERATOR_FOR_LINE3D_CLASS = ENABLE_MULOPERATOR_FOR_GEOMETRYCOMPUTATION
ENABLE_MULOPERATOR_FOR_PLANE3D_CLASS = ENABLE_MULOPERATOR_FOR_GEOMETRYCOMPUTATION

#-- global functions

def arange2(x, y, step=None):
	if (step == None):
		return np.arange(x, y + 1)
	else:
		return np.arange(x, y + step, step)

def is_scalar(x):
	return isinstance(x, numbers.Number)

def is_geometric_primitive(x):
	ty = type(x)
	flag = False
	for object in [point3d, line3d, plane3d]:
		flag |= ty is object
	return flag

def is_geometric(x):
	return is_geometric_primitive(x)

def ndim(x):
	return np.ndim(x)

def shape(x):
	return np.shape(x)

def sq(x):
	return x * x

def sqrt(x):
	return np.sqrt(x)

def rad2deg(x):
	#return (x * 180.0 / np.pi)
	return math.degrees(x)

def deg2rad(x):
	#return (x * np.pi / 180.0)
	return math.radians(x)

def rad2deg4list(radlist):
	deglist = []
	for i in range(len(radlist)):
		deglist.append(rad2deg(radlist[i]))
	return deglist

def deg2rad4list(deglist):
	radlist = []
	for i in range(len(deglist)):
		radlist.append(deg2rad(deglist[i]))
	return radlist

def sigmoid(x, tau=1.0, x0=0.0):
	return 1.0 / (1.0 + np.exp(-(x - x0) / tau))

def vecnorm(v):
	if (type(v) is not np.ndarray):
		v = np.array(v)
	return np.linalg.norm(v)

def norm(x):
	if (type(x) is np.ndarray):
		if (x.ndim == 1):
			return vecnorm(x)
		else:
			___ERROR___
	else:
		___ERROR___

def vecnormalize(v):
	ty = type(v)
	if (ty is list or ty is tuple):
		v = np.array(v)
	vlen = vecnorm(v)
	if (vlen != 0):
		v = v / vlen   # np.array の要素に対して、v /= vlen はできない
	if (ty is np.ndarray):
		pass
	elif  (ty is list):
		v = [v[0], v[1], v[2]]
	elif  (ty is tuple):
		v = (v[0], v[1], v[2])
	return v

def normalize(x):
	if (type(x) is np.ndarray):
		if (x.ndim == 1):
			return vecnormalize(x)
		else:
			___ERROR___
	else:
		___ERROR___

def vecangle(basev, v):
	u = basev
	nu = vecnorm(u)
	nv = vecnorm(v)
	if (nu == 0.0):
		return 0.0
	if (nv == 0.0):
		return 0.0
	u = u / nu # normalize
	v = v / nv # normalize
	c = np.inner(u, v)
	s = np.linalg.norm(np.cross(u, v)) # np.cross() は、ベクトルが両方とも2次元の場合はスカラー返すが、3次元の場合は3次元ベクトルを返す。
	return np.arctan2(s, c)

def angle(base, target):
	if (base is None or target is None):
		___ERROR___
	if (type(base) is list or type(base) is tuple):
		base = np.array(base)
	if (type(target) is list or type(target) is tuple):
		base = np.array(target)
	if (type(base) is np.ndarray and type(target) is np.ndarray):
		return vecangle(base, target)
	elif (type(base) is vec3d and type(target) is vec3d):
		return vecangle(base.get(), target.get())
	elif (type(base) is line3d and type(target) is line3d):
		return vecangle(base.getVector(), target.getVector())
	elif (type(base) is plane3d and type(target) is plane3d):
		return vecangle(base.getNormal(), target.getNormal())
	elif (type(base) is plane3d and type(target) is line3d):
		return np.pi / 2.0 - vecangle(base.getNormal(), target.getVector())
	elif (type(base) is line3d and type(target) is plane3d):
		return np.pi / 2.0 - vecangle(base.getVector(), target.getNormal())
	else:
		___ERROR___

def normalized(x):
	return x.normalized()

def inversed(x):
	return x.inversed()

def transposed(x):
	return x.transposed()

def conjugated(x):
	return x.conjugated()

def decimal_unit(x):
	if (x == 0.0):
		return 0.0
	elif (x > 0.0):
		sign = 1.0
	else:
		sign = -1.0
		x = -x
	unit = 1.0
	if (x > 1.0):
		while (unit < x):
			unit *= 10.0
		if (unit > x):
			unit /= 10.0
	else: # 0.0 < x < 1.0
		while (unit > x):
			unit /= 10.0
	return sign * unit

def logticks(min, max, ticks=10):
	ldprint2(["--> logticks({0}, {1}, {2})".format(min, max, ticks)])
	if (min < 0.0 or max < 0.0):
		print_error("Illegal arguments.")
		return None
	if (min > max):
		print_error("Illegal arguments.")
		return None
	if (ticks < 0):
		print_error("Illegal ticks.")
		return None
	tick_interval = 1.0 / float(ticks)
	ldprint3(["Tick interval: {0}".format(tick_interval)])
	min_log10 = np.log10(min)
	max_log10 = np.log10(max)
	unit_min_log10 = int(np.ceil(min_log10 * ticks))
	unit_max_log10 = int(np.floor(max_log10 * ticks))
	ldprint3(["min: {0} -> log10(min): {1} -> {2}".format(min, min_log10, unit_min_log10)])
	ldprint3(["max: {0} -> log10(max): {1} -> {2}".format(max, max_log10, unit_max_log10)])
	xlist = []
	if (min_log10 * ticks < unit_min_log10):
		xlist.append(min)
	for unit_tick_log10 in range(unit_min_log10, unit_max_log10 + 1):
		tick_log10 = unit_tick_log10 * tick_interval
		ldprint3(["Tick_log10: {0}".format(tick_log10)])
		tick = np.power(10.0, tick_log10)
		xlist.append(tick)
	if (max_log10 * ticks > unit_max_log10):
		xlist.append(max)
	ldprint(["xlist: ", xlist])
	ldprint2(["<-- logticks()"])
	return xlist

def pseudo_logticks(min, max, tick_scale=1.0):
	ldprint2(["--> pseudo_logticks({0}, {1}, {2})".format(min, max, tick_scale)])
	if (min < 0.0 or max < 0.0):
		print_error("Illegal arguments.")
		return None
	if (min > max):
		print_error("Illegal arguments.")
		return None
	if (tick_scale <= 0.0 or tick_scale > 1.0):
		print_error("Illegal tick scale.")
		return None
	min_unit = decimal_unit(min)
	max_unit = decimal_unit(max)
	if (max_unit < max):
		max_unit *= 10.0
	ldprint3(["unit: ", min, " -> ", min_unit])
	ldprint3(["unit: ", max, " -> ", max_unit])
	min_log10 = int(np.log10(min_unit))
	max_log10 = int(np.log10(max_unit))
	ldprint3(["log10: ", min_unit, " -> ", min_log10])
	ldprint3(["log10: ", max_unit, " -> ", max_log10])
	xlist = []
	for i in range(min_log10, max_log10):
		ldprint3(["log10: ", i])
		unit = np.power(10.0, i)
		ldprint3(["unit: ", unit])
		j_max = int(10.0 / tick_scale)
		for j in range(1, j_max):
			val = unit * tick_scale * j
			if (val < min):
				hval = unit * tick_scale * (j + 1)
				if (min < hval):
					xlist.append(min)
			elif (val < max):
				xlist.append(val)
			else:
				xlist.append(max)
				break
		if (val < max and max <= np.power(10.0, i + 1)):
			xlist.append(max)
	"""
	xlist = list(set(xlist))
	xlist = xlist.sort()
	"""
	ldprint2(["xlist: ", xlist])
	ldprint2(["<-- pseudo_logticks()"])
	return xlist

def _mat_4x4to3x3(m4):
	ty = type(m4)
	if (ty is cstrans3d):
		m43 = m4.getMatrix()
		ty = type(m4)
	if (ty is list):
		return [[m4[0, 0], m4[0, 1], m4[0, 2]],
		        [m4[1, 0], m4[1, 1], m4[1, 2]],
		        [m4[2, 0], m4[2, 1], m4[2, 2]]]
	elif (ty is tuple):
		return ((m4[0, 0], m4[0, 1], m4[0, 2]),
		        (m4[1, 0], m4[1, 1], m4[1, 2]),
		        (m4[2, 0], m4[2, 1], m4[2, 2]))
	elif (ty is np.ndarray):
		return np.array([[m4[0, 0], m4[0, 1], m4[0, 2]],
		                 [m4[1, 0], m4[1, 1], m4[1, 2]],
		                 [m4[2, 0], m4[2, 1], m4[2, 2]]])
	else:
		___ERROR___

def to3x3(m4):
	return _mat_4x4to3x3(m4)

def _mat_3x3to4x4(m3):
	ty = type(m3)
	if (ty is csrot3d):
		m3 = m3.getMatrix()
		ty = type(m3)
	if (ty is list):
		m = [[m3[0, 0], m3[0, 1], m3[0, 2], 0.0],
		     [m3[1, 0], m3[1, 1], m3[1, 2], 0.0],
		     [m3[2, 0], m3[2, 1], m3[2, 2], 0.0],
		     [0.0, 0.0, 0.0, 1.0]]
	elif (ty is tuple):
		m = ((m3[0, 0], m3[0, 1], m3[0, 2], 0.0),
		     (m3[1, 0], m3[1, 1], m3[1, 2], 0.0),
		     (m3[2, 0], m3[2, 1], m3[2, 2], 0.0),
		     (0.0, 0.0, 0.0, 1.0))
	elif (ty is np.ndarray):
		m = np.array([[m3[0, 0], m3[0, 1], m3[0, 2], 0.0],
		              [m3[1, 0], m3[1, 1], m3[1, 2], 0.0],
		              [m3[2, 0], m3[2, 1], m3[2, 2], 0.0],
		              [0.0, 0.0, 0.0, 1.0]])
	else:
		___ERROR___
	return m

def to4x4(m3):
	return _mat_3x3to4x4(m3)

def tovec3(v):
	return v[0:3]

def vec3(v):
	return tovec3(v)

def tovec4(v):
	dprint2('--> nkj.math.tovec4')
	dprint2('v: {0}'.format(v))
	if (len(v) == 3):
		if (type(v) is np.ndarray):
			v = np.append(v, 1.0)
		elif (type(v) is list or type(v) is tuple):
			v = v.append(1.0)
		else:
			___ERROR___
	elif (len(v) == 4):
		if (type(v) is np.ndarray):
			pass
		else:
			v = np.array(v)
	else:
		___ERROR___
	dprint2('v: {0}'.format(v))
	dprint2('<-- nkj.math.tovec4')
	return v

def vec4(v):
	return tovec4(v)

def offsetTransform(r, base):
	if (type(r) is np.ndarray):
		r = to4x4(r)
	else:
		return base @ r @ base.inv

def offset_transform(r, base):
	return offsetTransformation(r, base)

def offset(a, base):
	ty = type(a)
	if (ty is np.ndarray):
		sh = a.shape
		if (sh == (4, 4) or sh == (3, 4)):
			return offsetTransformation(a, base)
		else:
			___ERROR___
	elif (ty is list or ty is tuple):
		sh = shape(a)
		___NOT_IMPLEMENTED___
	elif (ty is trans3d):
		return offsetTransformation(a, base)
	else:
		___ERROR___

#-- classes

class complex():
	_classname = "nkj.math.complex"
	
	def __init__(self, x, y):
		#-- instance variables
		self._x = x
		self._y = y
	
	def __str__(self):
		return ns.concat(["(x, y): (", str(self.re()), ", ", str(self.im()), ")"])

	def __not__(self):
		return complex(self.re(), -self.im())

	def __add__(self, second):
		return complex(self.re() + second.re(), self.im() + second.im())

	def __sub__(self, second):
		return complex(self.re() - second.re(), self.im() - second.im())

	def __mul__(self, second): # 複素共役を掛け算
		ldprint(["self:   ", str(self)])
		ldprint(["second: ", str(second)])
		return self.re() * second.re() + self.im() * second.im()

	def __inv__(self): # '~'
		return self.normalized()

	def __getitem__(self, i):
		if (i == 0):
			return self._x
		elif (i == 1):
			return self._y
		else:
			return ValueError

	def __setitem__(self, i, second):
		if (i == 0):
			self._x = second
		elif (i == 1):
			self._y = second
		else:
			None

	@classmethod
	def getClassName(cls):
		return complex._classname

	@classmethod
	@property
	def classname(cls):
		return cls.getClassName()

	def set(self, x, y):
		self._x = x
		self._y = y

	def re(self):
		return self._x

	def im(self):
		return self._y

	def conjugated(self):
		return complex(self.re(), -self.im())

	def conj(self):
		return self.conjugated()

	def norm(self):
		return sq(self.re()) + sq(self.im())

	def normalized(self):
		norm = self.norm()
		if (norm == 0.0):
			return ValueError
		else:
			return complex(self.re() / norm, self.im() / norm)

	def phase(self): # 戻り値の範囲は [-pi, pi]
		return np.arctan2(self.im(), self.re())

#--- geometry

_DEFAULT_ROT2D = 0.0   # degree

class csrot2d():
	_classname = 'nkj.math.csrot2d'

	def __init__(self, a=None):
		self._a = a

_DEFAULT_QUAT = np.array([1.0, 0.0, 0.0, 0.0])   # (w, x, y, z)
_DEFAULT_CSROT3D = _DEFAULT_QUAT

class csrot3d():
	_classname = 'nkj.math.csrot3d'

	def __init__(self, a=None, flag=None, axis=None):
		if (a is None):
			self.set(_DEFAULT_CSROT3D)
		else:
			self.set(a, flag, axis)

	def __str__(self, title=None):
		return self.getPrintStr(title)

	def __mul__(self, second):
		return self.getMultipliedLocal(second)

	def __rmul__(self, first):
		return self.getMultipliedGlobal(first)

	def __matmul__(self, second):
		return self.__mul__(second)

	def __rmatmul__(self, first):
		return self.__rmul__(first)

	def __invert__(self):
		return self.getInversed()

	@classmethod
	def getClassName(cls):
		return quat._classname

	@classmethod
	@property
	def classname(cls):
		return cls.getClassName()

	def get(self):
		return self.array()

	def getAngle(self):
		return self.R.magnitude()

	def getAxis(self):
		return vecnormalize(self.R.as_rotvec())

	def getInversed(self):
		return csrot3d(self._r.inv())

	def getConjugated(self):
		return csrot3d(self.npquat.conjugate())

	def getMultipliedScalar(self, value):
		return carot3d(value * self.as_rotvec(), 'rotvec')

	def getMultipliedLocal(self, second):
		ldprint(["--> getMultipliedLocal()"])
		if (type(second) is csrot3d):
			return csrot3d(self._r @ second._r)
		elif (type(second) is R):
			return csrot3d(self._r @ second)
		elif (type(second) is list or type(second) is tuple):
			if (ndim(second) == 2):
				if (shape(second) == (3, 3)):
					return self.getMultipliedLocal(csrot3d(second))
				else:
					___ERROR___
			else:
				___ERROR___
		elif (type(second) is np.ndarray):
			if (second.dim == 2):
				if (second.shape == (3, 3)):
					return self.getMultipliedLocal(csrot3d(second))
				else:
					___ERROR___
			else:
				___ERROR___
		elif (is_scalar(second)):
			return getMultipliedScalar(second)
		else:
			___ERROR___

	def multipliedLocal(self, second):
		return self.getMultipliedLocal(second)

	def getMultipliedGlobal(self, first):
		if (type(first) is csrot3d):
			return csrot3d(first._r @ self._r)
		elif (type(first) is R):
			return csrot3d(first @ self._r)
		elif (type(first) is list or type(first) is tuple):
			if (ndim(first) == 2):
				if (shape(first) == (3, 3)):
					return self.getMultipliedGlobal(csrot3d(first))
				else:
					___ERROR___
			else:
				___ERROR___
		elif (type(first) is np.ndarray):
			if (first.dim == 2):
				if (first.shape == (3, 3)):
					return self.getMultipliedGlobal(csrot3d(first))
				else:
					___ERROR___
			else:
				___ERROR___
		elif (is_scalar(first)):
			return getMultipliedScalar(first)
		else:
			___ERROR___

	def multipliedGlobal(self, second):
		return self.getMultipliedGlobal(second)

	def getRotatedLocal(self, r, flag=None):
		return self.getMultiplinedLocal(csrot3d(r, flag))

	def rotatedLocal(self, r, flag=None):
		return self.getRotatedLocal(r, flag)

	def getRotatedGlobal(self, r, flag=None):
		return self.getMultiplinedGlobal(csrot3d(r, flag))

	def rotatedGlobal(self, second):
		return self.getRotatedGlobal(second)

	def getPrintStr(self, title=None):
		if (title is None):
			mesg = ''
		else:
			mesg = '{0}: '.format(title)
		q = self.quat
		w = q[0]; x = q[1]; y = q[2]; z = q[3]
		c = w
		s = np.sqrt(x**2 + y**2 + z**2)
		angle = 2.0 * np.arctan2(s, c)
		if (s == 0.0):
			rotaxis = (x, y, z)
		else:
			rotaxis = (x / s, y / s, z / s)
		mesg += 'angle: {0:.3f} deg, '.format(rad2deg(angle))
		mesg += 'axis: ({0:.3f}, {1:.3f}, {2:.3f})'.format(rotaxis[0], rotaxis[1], rotaxis[2])
		return mesg

	def _normalize(self, a):
		mag = np.sqrt(a[0]**2 + a[1]**2 + a[2]**2 + a[3]**2)
		if (mag == 0.0):
			return a
		else:
			return a / mag 

	def _array_np2r(self, a):
		a = self._normalize(a)
		return [a[1], a[2], a[3], a[0]] # (w, x, y, z) to (x, y, z, w)

	def _array_r2np(self, a):
		a = self._normalize(a)
		return [a[3], a[0], a[1], a[2]] # (x, y, z, w) to (w, x, y, z)

	def set(self, a, flag=None, axis=None):
		ldprint('--> csrot3d.set()')
		ty = type(a)
		if (ty is csrot3d):
			self._r = a._r
		elif (ty is R):
			self._r = a
		elif (ty is list or ty is tuple):
			dim = ndim(a)
			if (dim == 1):
				sh = shape(a)
				if (sh == (4,)):
					if (flag is None or flag == 'quaternion' or flag == 'quat'):
						self._r = R.from_quat(self._array_np2r(a)) # quaternion
					elif (flag == 'R.quaternion' or flag == 'rquat'):
						self._r = R.from_quat(copy.copy(a))
					else:
						___ERROR___
				elif (sh == (3,)):
					if (flag is None or flag == 'rotationalvector' or flag == 'rotvec'):
						self._r = R.from_rotvec(copy.copy(a))
					elif (flag == 'directioncosines' or flag == 'dircos'):
						if (axis is None):
							axis = np.array([0, 0, 1])
						x = vecnormalize(copy.copy(a))
						axis = vecnormalize(axis)
						raxis = np.cross(axis, x)
						angle = vecangle(axis, x)
						rvec = angle * raxis
						self.set(rvec, 'rotvec')
					elif (flag == 'directionangles' or flag == 'dirang'):
						r = [np.cos(a[0]), np.cos(a[1]), np.cos(a[2])]
						self.set(r, 'dircos')
					elif (flag == 'euler_zyx'):
						self._r = R.from_euler('zyx', copy.copy(a)) # extrinsic rotation
					elif (flag == 'euler_ZYX'):
						self._r = R.from_euler('ZYX', copy.copy(a)) # intrinsic rotation
					elif (flag == 'euler_xyx'):
						self._r = R.from_euler('xyx', copy.copy(a))
					elif (flag == 'euler_XYX'):
						self._r = R.from_euler('XYX', copy.copy(a))
					else:
						___ERROR___
				elif (sh == (2,)):
					if (flag is None or flag == 'anglevec' or flag == 'angleaxis'):
						angle = a[0]
						if (not is_scalar(angle)):
							___ERROR___
						vec = a[1]
						if (shape(vec) == (3,)):
							rvec = angle * vecnormalize(np.array(vec))
							self.set(rvec, 'rotvec')
						else:
							___ERROR___
					else:
						___ERROR___
				else:
					___ERROR___
			elif (dim == 2):
				sh = shape(a)
				if (sh == (3, 3)):
					if (flag is None or flag == 'matrix' or flag == 'mat'):
						self._r = R.from_matrix(copy.copy(a))
					else:
						___ERROR___
				else:
					___ERROR___
			else:
				___ERROR___
		elif (ty is np.ndarray):
			dim = a.ndim
			if (dim == 1):
				sh = a.shape
				if (sh == (4,)):
					if (flag is None or flag == 'quaternion' or flag == 'quat'):
						self._r = R.from_quat(self._array_np2r(a))
					elif (flag == 'R.quaternion' or flag == 'rquat'):
						self._r = R.from_quat(copy.copy(a))
					else:
						___ERROR___
				elif (sh == (3,)):
					if (flag is None or flag == 'rotationalvector' or flag == 'rotvec'):
						self._r = R.from_rotvec(copy.copy(a))
					elif (flag == 'euler_zyx'):
						self._r = R.from_euler('zyx', copy.copy(a)) # extrinsic rotation
					elif (flag == 'euler_ZYX'):
						self._r = R.from_euler('ZYX', copy.copy(a)) # intrinsic rotation
					elif (flag == 'euler_xyx'):
						self._r = R.from_euler('xyx', copy.copy(a))
					elif (flag == 'euler_XYX'):
						self._r = R.from_euler('XYX', copy.copy(a))
					else:
						___ERROR___
				elif (sh == (2,)):
					if (flag is None or flag == 'anglevec' or flag == 'angleaxis'):
						angle = a[0]
						if (not is_scalar(angle)):
							___ERROR___
						vec = a[1]
						if (shape(vec) == (3,)):
							rvec = angle * vecnormalize(np.array(vec))
							self.set(rvec, 'rotvec')
						else:
							___ERROR___
					else:
						___ERROR___
				else:
					___ERROR___
			elif (dim == 2):
				sh = a.shape
				if (sh == (3, 3)):
					self._r = R.from_matrix(copy.deepcopy(a)) # 3x3 matrix
				elif (sh == (4, 4)):
					self._r = R.from_matrix(_mat_4x4to3x3(a)) # 4x4 matrix
				else:
					___ERROR___
			else:
				___ERROR___
		else:
			___ERROR___
		ldprint('<-- csrot3d.set()')

	def getR(self):
		return self._r

	@property
	def R(self):
		return self.getR()

	def setR(self, r):
		self.set(r)

	@R.setter
	def R(self, second):
		self.setR(second)

	@property
	def rquatarray(self):
		return self._r.as_quat()

	@rquatarray.setter
	def rquatarray(self, second):
		set.set(second)

	@property
	def rquat(self):
		return self.rquatarray

	@rquat.setter
	def rquat(self, second):
		self.set(second)

	@property
	def npquatarray(self):
		return self._array_r2np(self._r.as_quat())

	@npquatarray.setter
	def npquatarray(self, second):
		self.set(second, 'np.quaternion')

	@property
	def quat(self):
		return self.rquatarray

	@quat.setter
	def quat(self, second):
		self.set(second, 'quaternion')

	def getMatrix(self):
		return self._r.as_matrix()

	@property
	def matrix(self):
		return self.getMatrix()

	@property
	def mat(self):
		return self.getMatrix()

	@property
	def m(self):
		return self.getMatrix()

	def setMatrix(self, second):
		self.set(second)

	@matrix.setter
	def matrix(self, second):
		self.setMatrix(second)

	@mat.setter
	def mat(self, second):
		self.setMatrix(second)

	@m.setter
	def m(self, second):
		self.setMatrix(second)

	@property
	def rotvec(self):
		return self._r.as_rotvec()

	@rotvec.setter
	def rotvec(self, second):
		self.set(second, 'rotvec')

	@property
	def mrp(self): # Modified Rodrigues Parameters (MRPs)
		return self._r.as_mrp()

	@mrp.setter
	def mrp(self, second):
		self.set(second, 'mrp')

	def euler(self, flag=None):
		if (flag is None):
			return self._r.as_euler()
		else:
			return self._r.as_euler(flag)

	@property
	def w(self):
		return self.rquat[3]

	@property
	def x(self):
		return self.rquat[0]

	@property
	def y(self):
		return self.rquat[1]

	@property
	def z(self):
		return self.rquat[2]

	@property
	def rx(self):
		return self.rotvec[0]

	@property
	def ry(self):
		return self.rotvec[1]

	@property
	def rz(self):
		return self.rotvec[2]

	@property
	def angle(self):
		return self.getAngle()

	@property
	def ang(self):
		return self.angle

	@property
	def axis(self):
		return self.getAxis()

	@property
	def vec(self):
		return self.axis

	@property
	def v(self):
		return self.axis

	@property
	def inversed(self):
		return self.getInversed()

	@property
	def inv(self):
		return self.inversed

	@property
	def i(self):
		return self.inversed

	def inverse(self):
		self.set(self.getInversed())

	@property
	def conjugated(self):
		return self.getConjugated()

	@property
	def conj(self):
		return self.conjugated

	@property
	def c(self):
		return self.conjugated

	def conjugate(self):
		self.set(self.GetConjugated())

	def rorated(self, a):
		return self.getRotated()

	def rorate(self, a):
		self.set(self.rorated(a))

	def print(self, title=None):
		print(self.getPrintStr(title))

_DEFAULT_CSTRANS3D = [np.array([1.0, 0.0, 0.0, 0.0]), np.array([0.0, 0.0, 0.0])]

class cstrans3d():
	_classname = 'nkj.math.cstrans3d'

	def __init__(self, a=None, flag=None):
		if (a is None):
			self.set(_DEFAULT_CSTRANS3D)
		else:
			self.set(a, flag)

	def __str__(self, title=None):
		return self.getPrintStr(title)

	def __matmul__(self, second):
		return self.getMultipliedGlobal(second)

	def __rmatmul__(self, first):
		return self.getMultipliedLocal(first)

	if (ENABLE_MULOPERATOR_FOR_CSTRANS3D_CLASS):
		def __mul__(self, second):
			return self.__matmul__(second)

		def __rmul__(self, first):
			return self.__rmatmul__(first)

	def __invert__(self):
		return self.getInversed()

	@classmethod
	def getClassName(cls):
		return cstrans3d._classname

	@classmethod
	@property
	def classname(cls):
		return cls.getClassName()

	def getMultipliedScalar(self, value):
		r = value * self._r.rotvec
		t = value * self._t
		return cstrans3d([r.quat, t])

	def multipliedScalar(self, value):
		return self.getMultipliedScalar(value)

	def getMultipliedGlobal(self, second):
		ty = type(second)
		if (ty is csrot3d):
			return cstrans3d(self.matrix @ to4x4(second.getMatrix()))
		elif (ty is cstrans3d):
			return cstrans3d(self.matrix @ second.getMatrix())
		elif (ty is list or ty is tuple):
			if (ndim(second) == 2):
				m = second
				if (shape(m) == (3, 3)):
					return self.getMultipliedGlobal(csrot3d(m))
				elif (shape(m) == (4, 4)):
					return self.getMultipliedGlobal(cstrans3d(m))
				else:
					___ERROR___
			elif (ndim(second) == 1): # vector
				v = second
				return self.getMultipliedGlobal(np.array(v, dtype=np.float32))
			else:
				___ERROR___
		elif (ty is np.ndarray):
			if (second.ndim == 2):
				m = second
				if (m.shape == (3, 3)):
					return self.getMultipliedGlobal(csrot3d(m))
				elif (m.shape == (4, 4)):
					return self.getMultipliedGlobal(cstrans3d(m))
				else:
					___ERROR___
			elif (second.ndim == 1): # vector
				v = second
				if (v.shape == (3,)):
					return tovec3(self.getMultipliedGlobal(tovec4(v)))
				elif (v.shape == (4,)):
					m = self.matrix
					return m @ v
				else:
					___NOT_IMPLEMENTED___
			else:
				___ERROR___
		elif (is_scalar(second)):
			return getMultipliedScalar(second)
		elif (is_geometric(second)):
			return second.getMultiplied(self)
		else:
			___ERROR___

	def multipliedGlobal(self, second):
		return getMultipliedGlobal(second)

	def getMultipliedLocal(self, first):
		ty = type(first)
		if (ty is csrot3d):
			return cstrans3d(to4x4(first.getMatrix()) @ self.matrix)
		elif (ty is cstrans3d):
			return cstrans3d(first.getMatrix() @ self.matrix)
		elif (ty is list or ty is tuple):
			if (ndim(first) == 2):
				m = first
				if (shape(m) == (3, 3)):
					return self.getMultipliedLocal(csrot3d(m))
				elif (shape(m) == (4, 4)):
					return self.getMultipliedLocal(cstrans3d(m))
				else:
					___ERROR___
			elif (ndim(first) == 1): # vector
				v = first
				return self.getMultipliedLocal(np.array(v, dtype=np.float32))
			else:
				___ERROR___
		elif (ty is np.ndarray):
			if (first.ndim == 2):
				m = first
				if (m.shape == (3, 3)):
					return self.getMultipliedLocal(csrot3d(m))
				elif (m.shape == (4, 4)):
					return self.getMultipliedLocal(cstrans3d(m))
				else:
					___ERROR___
			elif (first.ndim == 1): # vector
				v = first
				return self.getMultipliedGlobal(v) # The same value is returned.
			else:
				___ERROR___
		elif (is_scalar(secon)):
			return getMultipliedScalar(secon)
		elif (is_geometric(secon)):
			return first.getMultiplied(self)
		else:
			___ERROR___

	def multipliedLocal(self, first):
		return getMultipliedLocal(first)

	def getPrintStr(self, title=None):
		if (title is None):
			mesg = ''
		else:
			mesg = '{0}: '.format(title)
		mesg += 'r:('
		mesg += self._r.getPrintStr()
		mesg += '), t:({0:.3f}, {1:.3f}, {2:.3f})'.format(self._t[0], self._t[1], self._t[2])
		return mesg

	def set(self, a, flag=None):
		ldprint(['--> set({0}, \'{1}\')'.format(a, flag)])
		ty = type(a)
		if (ty is cstrans3d):
			self._r = a._r
			self._t = a._t
			if (flag is not None):
				___ERROR___
		elif (ty is list or ty is tuple):
			sh = shape(a)
			if (sh == (2,) or sh == (2, 3)):
				r = a[0]
				t = a[1]
				ldprint("r: {0}".format(r))
				ldprint("t: {0}".format(t))
				self._r = csrot3d(r, flag)
				self._t = np.array([t[0], t[1], t[2]])
			else:
				___ERROR___
		elif (ty is np.ndarray):
			sh = a.shape
			if (sh == (3, 3)):
				self._r = csrot3d([[a[0, 0], a[0, 1], a[0, 2]],
				                   [a[1, 0], a[1, 1], a[1, 2]],
				                   [a[2, 0], a[2, 1], a[2, 2]]])
			elif (sh == (3, 4) or sh == (4, 4)):
				self._r = csrot3d([[a[0, 0], a[0, 1], a[0, 2]],
				                   [a[1, 0], a[1, 1], a[1, 2]],
				                   [a[2, 0], a[2, 1], a[2, 2]]])
				self._t = np.array([a[0, 3], a[1, 3], a[2, 3]])
		else:
			___ERROR___

	@property
	def R(self):
		return self._r.R

	@R.setter
	def R(self, second):
		self._r.set(R)

	@property
	def rquatarray(self):
		return self._r.rquatarray

	@rquatarray.setter
	def rquatarray(self, second):
		self._r.set(second)

	@property
	def rquat(self):
		return self.rquatarray

	@rquat.setter
	def rquat(self, second):
		self.set(second)

	@property
	def npquatarray(self):
		return self._r.npquatarray

	def getQuat(self):
		return self._r.quat

	@property
	def quat(self):
		return self.getQuat()

	def setQuat(self, second):
		self.set(second, 'quaternion')

	@quat.setter
	def quat(self, second):
		self.setQuat(second)

	def getRotationalMatrix(self):
		return self._r.matrix

	def getRotationMatrix(self):
		return self.getRotationalMatrix()

	def getRotMatrix(self):
		return self.getRotationalMatrix()

	def getRM(self):
		return self.getRotationalMatrix()

	@property
	def rotmatrix(self):
		return self.getRotMatrix()

	@property
	def rotmat(self):
		return self.getRotMatrix()

	@property
	def rmat(self):
		return self.getRotMatrix()

	@property
	def rm(self):
		return self.getRotMatrix()

	def setRotationalMatrix(self, second):
		self.set(second)

	def setRotationMatrix(self, second):
		self.setRotationalMatrix(second)

	def setRotMatrix(self, second):
		self.setRotationalMatrix(second)

	def setRM(self, second):
		self.setRotationalMatrix(second)

	@rotmatrix.setter
	def rotmatrix(self, second):
		self.setRotationalMatrix(second)

	@rotmat.setter
	def rotmat(self, second):
		self.setRotationalMatrix(second)

	@rmat.setter
	def rmat(self, second):
		self.setRotationalMatrix(second)

	@rm.setter
	def rm(self, second):
		self.setRotationalMatrix(second)

	def getMatrix(self):
		m3 = self._r.matrix
		m4 = np.array(to4x4(m3))
		t = self._t
		m4[0, 3] = t[0]
		m4[1, 3] = t[1]
		m4[2, 3] = t[2]
		return m4

	@property
	def matrix(self):
		return self.getMatrix()

	@property
	def setRotMatrix(self, second):
		self.setRotationalMatrix(second)

	@rotmatrix.setter
	def rotmatrix(self, second):
		self.setRotMatrix(second)

	@rotmat.setter
	def rotmat(self, second):
		self.setRotMatrix(second)

	@rmat.setter
	def rmat(self, second):
		self.setRotMatrix(second)

	@rm.setter
	def rm(self, second):
		self.setRotMatrix(second)

	def getMatrix(self):
		m3 = self._r.matrix
		m4 = np.array(to4x4(m3))
		t = self._t
		m4[0, 3] = t[0]
		m4[1, 3] = t[1]
		m4[2, 3] = t[2]
		return m4

	@property
	def matrix(self):
		return self.getMatrix()

	@property
	def mat(sef):
		return self.getMatrix()

	@property
	def m(sef):
		return self.getMatrix()

	def setMatrix(self, second):
		self.set(second)

	@matrix.setter
	def matrix(self, second):
		self.setMatrix(second)

	@mat.setter
	def mat(self, second):
		self.setMatrix(second)

	@m.setter
	def m(self, second):
		self.setMatrix(second)

	def getRotvec(self):
		return self._r.as_rotvec()

	@property
	def rotvec(self):
		return self.getRotvec()

	@property
	def rvec(self):
		return self.getRotvec()

	def setRotvec(self, second):
		self.set(second, 'rotvec')

	@rotvec.setter
	def rotvec(self, second):
		self.setRotvec(second)

	@rvec.setter
	def rvec(self, second):
		self.setRotvec(second)

	def getAxes(self):
		rm = self.getRotationalMatrix()
		xaxis = rm[:, 0]
		yaxis = rm[:, 1]
		zaxis = rm[:, 2]
		return xaxis, yaxis, zaxis

	@property
	def axes(self):
		return self.getAxes()

	def setAxes(self, xaxis, yaxis, zaxis):
		rm = np.zeros((3, 3), dtype=np.float32)
		rm[:, 0] = xaxis
		rm[:, 1] = yaxis
		rm[:, 2] = zaxis
		ldprint(f'xaxis: {xaxis}')
		ldprint(f'yaxis: {yaxis}')
		ldprint(f'zaxis: {zaxis}')
		ldprint(f'{rm}')
		self.setRotationalMatrix(rm)

	@axes.setter
	def axes(self, axes):
		ldprint(f'xaxis: {axes[0]}')
		ldprint(f'yaxis: {axes[1]}')
		ldprint(f'zaxis: {axes[2]}')
		self.setAxes(axes[0], axes[1], axes[2])

	def getTranslation(self):
		return self._t

	def getTrans(self):
		return self.getTranslation()

	def getT(self):
		return self.getTranslation()

	def getOrigin(self):
		return self.getTranslation()

	def getOrig(self):
		return self.getOrigin()

	@property
	def translation(self):
		return self.getTranslation()

	@property
	def trans(self):
		return self.getTranslation()

	@property
	def t(self):
		return self.getTranslation()

	@property
	def origin(self):
		return self.getOrigin()

	@property
	def orig(self):
		return self.getOrigin()

	def setTranslation(self, second):
		self._t = np.array(second)

	def setTrans(self, second):
		self.setTranslation(second)

	def setT(self, second):
		self.setTranslation(second)

	def setOrigin(self, second):
		self.setTranslation(second)

	def setOrig(self, second):
		self.setOrigin(second)

	@translation.setter
	def translation(self, second):
		self.setTranslation(second)

	@trans.setter
	def trans(self, second):
		self.setTranslation(second)

	@t.setter
	def t(self, second):
		self.setTranslation(second)

	@origin.setter
	def origin(self, second):
		self.setOrigin(second)

	@orig.setter
	def orig(self, second):
		self.setOrigin(second)

	def getXaxis(self):
		return self.getRotMatrix()[:, 0]

	@property
	def xaxis(self):
		return getXaxis()
	
	def getYaxis(self):
		return self.getRotMatrix()[:, 1]

	@property
	def yaxis(self):
		return getYaxis()
	
	def getZaxis(self):
		return self.getRotMatrix()[:, 2]

	@property
	def zaxis(self):
		return getZaxis()
	
	def getInversed(self):
		return cstrans3d(np.linalg.inv(self.matrix))

	@property
	def inversed(self):
		return self.getInversed()

	@property
	def inv(self):
		return self.getInversed()

	@property
	def i(self):
		return self.getInversed()

	def inverse(self):
		self.set(self.getInversed())

	def getRotatedLocal(self, a, flag=None):
		return self.getMultipliedLocal(csrot3d(a, flag))

	def getRotatedGlobal(self, a, flag=None):
		return self.getMultipliedGlobal(to4x4(csrot3d(a, flag).getMatrix()))

	def getTranslatedLocal(self, a):
		t = copy.deepcopy(self)
		t.translateLocal(a)
		return t

	def getTranslatedGlobal(self, a):
		t = copy.deepcopy(self)
		t.translateGlobal(a)
		return t

	def rotateLocal(self, a, flag=None):
		self.set(self.getRotatedLocal(a, flag))

	def rotateGlobal(self, a, flag=None):
		self.set(self.getRotatedGlobal(a, flag))

	def rotate_local(self, a, flag=None):
		self.rotateLocal(a, flag)

	def rotate_global(self, a, flag=None):
		self.rotateGlobal(a, flag)

	def rotate(self, a, flag=None):
		self.rotate_local(a, flag)

	def translateLocal(self, a):
		dprint2('--> {0}.translateLocal()'.format(self.getClassName()))
		dprint2('a:    {0}'.format(a))
		dprint2('vec4: {0}'.format(tovec4(a)))
		dprint2('-- m --\n{0}\n--\n'.format(self.getMatrix()))
		dprint2('dot:  {0}'.format(self.getMatrix().dot(tovec4(a))))
		self.setTranslation(self.getTranslation() + tovec3(self.getMatrix().dot(tovec4(a))))
		dprint2('<-- {0}.translateLocal()'.format(self.getClassName()))

	def translateGlobal(self, a):
		self.setTranslation(self.getTranslation() + tovec3(a))

	def translate_local(self, a):
		self.translateLocal(self, a)

	def translate_global(self, a):
		self.translateGlobal(self, a)

	def translate(self, a):
		self.translate_local(a, flag)

	def print(self, title=None):
		print(self.getPrintStr(title))

_DEFAULT_MAT3X3 = np.identity(3)

class mat3x3():
	_classname = 'nkj.math.mat3x3'

	def __init__(self, a=None):
		if (a is None):
			self.set(_DEFAULT_MAT3X3)
		else:
			self.set(a)

	def __str__(self, title=None):
		if (title is None):
			mesg = "---\n"
		else:
			mesg = "--- {0} ---\n".format(title)
		mesg += "{0: .3f}, {1: .3f}, {2: .3f}\n".format(self._m[0][0], self._m[0][1], self._m[0][2])
		mesg += "{0: .3f}, {1: .3f}, {2: .3f}\n".format(self._m[1][0], self._m[1][1], self._m[1][2])
		mesg += "{0: .3f}, {1: .3f}, {2: .3f}\n".format(self._m[2][0], self._m[2][1], self._m[2][2])
		mesg += "---"
		return mesg

	@classmethod
	def getClassName(cls):
		return mat3x3._classname

	@classmethod
	@property
	def classname(cls):
		return cls.getClassName()

	def set(self, m):
		ldprint2(["--> set()"])
		if (type(a) is mat4x4):
			self._m = m._m
		elif (type(m) is np.ndarray):
			ldprint(["shape: {0}".format(m.shape)])
			if (m.shape == (3, 3)):
				self._m = copy.deepcopy(m) # copy を使わないと、元の配列書換えで要素が変更されてしまう。また、copy.copy()だとネストする配列には対応しないので、copy.deepcopy()を使う。
			elif (m.shape == (4,)): # 要素が'4'のみのタプル．要素が一つの場合、'(4)'ではなく、'(4,)'（カンマが必要）と書く。
				ldprint(["angle: {0:.3f}, vector: ({1: .3f}, {2: .3f}, {3: .3f})".format(rad2deg(m[0]), m[1], m[2], m[3])])
				"""
				self._m = 
				■実装すること
				"""
			else:
				___ERROR___
		else:
			___ERROR___
		ldprint2(["<-- set()"])

	def identity(self):
		self.set(np.identity(4))

	def inverse(self):
		self.set(np.linalg.inv(self.get()))

_DEFAULT_MAT4X4 = np.identity(4)

class mat4x4(): # python 3.5 より @ 演算子がサポートされたので、あえてこのクラスを使うメリットはない。
	_classname = 'nkj.math.mat4x4'
	
	def __init__(self, a=None):
		if (a is None):
			self.set(_DEFAULT_MAT4X4)
		else:
			self.set(a)

	def __str__(self, title=None):
		return getPrintStr(title)

	def __invert__(self): # '~'単項演算子
		return self.getInversed()

	def __matmul__(self, second):
		"""
		m = np.zeros((4, 4))
		for j in range(4):
			for i in range(4):
				for ii in range(4):
					m[j][i] = m[j][i] + self.array()[j][ii] * second.array()[ii][i]
		"""
		ldprint2(["second type: ", type(second)])
		if (type(second) is mat4x4):
			m = self.array() @ second.array() # np.matmul() あるいは np.dot() でも良い
		else:
			___ERROR___
		return m

	if (ENABLE_MULOPERATOR_FOR_MAT4X4_CLASS):
		def __mul__(self, second):
			return self.__matmul__(second)

	@classmethod
	def getClassName(cls):
		return mat4x4._classname

	@classmethod
	@property
	def classname(cls):
		return cls.getClassName()

	@property
	def dim(self):
		return self.array().ndim

	@property
	def size(self):
		return self.array().shape

	@property
	def array(self):
		return self._m

	def get(self):
		return self.array()

	def getInversed(self):
		return mat4x4(np.linalg.inv(self.get()))

	@property
	def inversed(self):
		return self.getInversed()

	@property
	def inv(self):
		return self.getInversed()

	@property
	def i(self):
		return self.getInversed()

	def getIdentity(self):
		return mat4x4(np.identity(4))

	@property
	def identity(self):
		return self.getIdentity()

	def getPrintStr(self, title=None):
		if (title is None):
			mesg = "---\n"
		else:
			mesg = "--- {0} ---\n".format(title)
		mesg += "{0: .3f}, {1: .3f}, {2: .3f}, {3: .3f}\n".format(self._m[0][0], self._m[0][1], self._m[0][2], self._m[0][3])
		mesg += "{0: .3f}, {1: .3f}, {2: .3f}, {3: .3f}\n".format(self._m[1][0], self._m[1][1], self._m[1][2], self._m[1][3])
		mesg += "{0: .3f}, {1: .3f}, {2: .3f}, {3: .3f}\n".format(self._m[2][0], self._m[2][1], self._m[2][2], self._m[2][3])
		mesg += "{0: .3f}, {1: .3f}, {2: .3f}, {3: .3f}\n".format(self._m[3][0], self._m[3][1], self._m[3][2], self._m[3][3])
		mesg += "---"
		return mesg

	@property
	def printstr(self):
		return self.getPrintStr()

	@property
	def str(self):
		return self.getPrintStr()

	def array(self):
		return self._m

	def set(self, m):
		ldprint2(["--> set()"])
		if (type(m) is mat4x4):
			self._m = m._m
		elif (type(m) is cstrans3d):
			self._m = m.matrix
		elif (type(m) is np.ndarray):
			ldprint(["shape: {0}".format(m.shape)])
			if (m.shape == (4, 4)):
				self._m = copy.deepcopy(m)
			elif (m.shape == (4,)): # 要素が'4'のみのタプル．要素が一つの場合、'(4)'ではなく、'(4,)'（カンマが必要）と書く。
				ldprint(["angle: {0:.3f}, vector: ({1: .3f}, {2: .3f}, {3: .3f})".format(rad2deg(m[0]), m[1], m[2], m[3])])
				"""
				self._m = 
				実装すること
				"""
				___ERROR___
			else:
				___ERROR___
		else:
			___ERROR___
		"""
		self._m = np.ndarray((4, 4))
		for j in range(4):
			for i in range(4):
				self._m[j][i] = m[j][i]
		"""
		ldprint2(["<-- set()"])

	def inverse(self):
		self.set(self.getInversed())

	def getRotation(self):
		return to3x3(self._m)

	def getColumn3(self, i):
		return self._m[:3, i]

	def getXaxis(self):
		return self.getColumn3(0)

	@property
	def xaxis(self):
		return self.getXaxis()

	def getYaxis(self):
		return self.getColumn3(0)

	@property
	def yaxis(self):
		return self.getYaxis()

	def getZaxis(self):
		return self.getColumn3(0)

	@property
	def zaxis(self):
		return self.getZaxis()

	def getTranslation(self):
		return self._m[:3, 3]

	def getTrans(self):
		return self.getTranslation()

	@property
	def translation(self):
		return self.getTranslation()

	@property
	def trans(self):
		return self.getTranslation()

	@property
	def t(self):
		return self.getTranslation()

	"""
	def rotate(self, angle, base=None):
		if (type(base) is np.ndarray):
		elif (base is None):
		else:
			___ERROR___
	"""

	def print(self, title=None):
		print(self.getPrintStr(title))

_DEFAULT_VEC3D = np.array([0.0, 0.0, 0.0])

class vec3d:
	_classname = 'nkj.math.vec3d'

	def __init__(self, v=None):
		if (v is None):
			self._v = _DEFAULT_VEC3D
		else:
			self.set(v)

	def __str__(self, title=None):
		return self.getPrintStr(title)

	def __pos__(self):
		return self

	def __neg__(self):
		return vec3d(-self.array())

	def __invert__(self):
		return self.__neg__()

	def __rmatmul__(self, first):
		return self.getMultiplied(first)

	if (ENABLE_MULOPERATOR_FOR_LINE3D_CLASS):
		def __rmul__(self, first):
			return self.__rmatmul__(first)

	@classmethod
	def getClassName(cls):
		return point3d._classname

	@classmethod
	@property
	def classname(cls):
		return cls.getClassName()

	def set(self, p):
		if (type(p) is vec3d):
			self._v = p._v
		elif (type(p) is np.ndarray):
			self._v = p
		elif (type(p) is list or type(p) is tuple):
			self._v = np.array(p)
		else:
			___ERROR___

	def get(self):
		return self._v

	def getArray(self):
		return np.array(self.get(), dtype=np.float32)

	@property
	def array(self):
		return self.getArray()

	def setArray(self):
		self.set(v)

	@array.setter
	def array(self, v):
		self.setArray(v)

	def getList(self):
		return list(self.getArray())

	@property
	def list(self):
		return self.getList()

	def getX(self):
		return self._v[0]

	def getY(self):
		return self._v[1]

	def getZ(self):
		return self._v[2]

	@property
	def x(self):
		return self.getX()

	@property
	def y(self):
		return self.getY()

	@property
	def z(self):
		return self.getZ()

	def setX(self, v):
		self._v[0] = v

	def setY(self, v):
		self._v[1] = v

	def setZ(self, v):
		self._v[2] = v

	@x.setter
	def x(self, v):
		return self.setX(v)

	@y.setter
	def y(self, v):
		return self.setY(v)

	@z.setter
	def z(self, v):
		return self.setZ(v)

	def getMultiplied(self, r):
		if (type(r) is csrot3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is cstrans3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is list or type(r) is tuple):
			return self.getMultiplied(np.array(r))
		elif (type(r) is np.ndarray):
			if (r.ndim == 2):
				if (r.shape == (3, 3)):
					v = r @ self._v
					return vec3d(v)
				elif (r.shape == (4, 4)):
					v = vec3(r @ vec4(self._v))
					return vec3d(v)
				else:
					___ERROR___
			else:
				___ERROR___
		else:
			___ERROR___

	def multiplied(self, r):
		return self.getMultiplied(r)

	def getRotated(self, r):
		if (type(r) is csrot3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is cstrans3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is list or type(r) is tuple):
			return self.getMultiplied(np.array(r))
		elif (type(r) is np.ndarray):
			if (r.shape == (3, 3)):
				v = r @ self._v
				return vec3d(v)
			elif (r.shape == (4, 4)):
				v = vec3(r @ vec4(self._v))
				return vec3d(v)
			else:
				___ERROR___
		else:
			___ERROR___

	def rotated(self, r):
		return getRotated(r)

	def getTranslated(self, t):
		if (type(t) is cstrans3d):
			return self.getTranslated(t.getTranslation())
		elif (type(t) is list or type(t) is tuple):
			return self.getTranslated(np.array(t))
		elif (type(t) is np.ndarray):
			v = self._v + t
			return vec3d(v)
		else:
			___ERROR___

	def translated(self, t):
			return self.getTranslated(t)

	def getTransformed(self, t):
		return self.getMultiplied(t)

	def transformed(self, t):
		return self.getTransformed(t)

	def getPrintStr(self, title=None):
		if (title is None):
			mesg = ''
		else:
			mesg = title + ': '
		mesg += "({0: .3f}, {1: .3f}, {2: .3f})".format(self._v[0], self._v[1], self._v[2])
		return mesg

	@property
	def printstr(self):
		return self.getPrintStr()

	@property
	def str(self):
		return self.getPrintStr()

	def print(self, title=None):
		print(self.getPrintStr(title))

	def getDataStr(self):
		return '{0:21.12f}, {1:21.12f}, {2:21.12f}\n'.format(self._v[0], self._v[1], self._v[2])

	@property
	def datastr(self):
		return self.getDataStr()

	def setDataStr(self, s):
		s = s.split(',')
		for i in range(len(s)):
			s[i] = float(s[i].strip())
		self.set([s[0], s[1], s[2]])

	@datastr.setter
	def datastr(self, s):
		return setDataStr(s)

	def load(self, filename):
		with open(filename, 'r') as f:
			self.setDataStr(f.read())

	def save(self, filename):
		with open(filename, 'w') as f:
			f.write(self.getDataStr())

class point3d(vec3d):
	_classname = 'nkj.math.point3d'

	def __init__(self, p=None):
		super().__init__(p)

	@classmethod
	def getClassName(cls):
		return point3d._classname

	def set(self, p):
		if (type(p) is point3d):
			super().set(p._v)
		else:
			super().set(p)

_DEFAULT_LINE3D_POINT1 = np.array([0.0, 0.0, 0.0])
_DEFAULT_LINE3D_POINT2 = np.array([0.0, 0.0, 1.0])

class line3d:
	_classname = 'nkj.math.line3d'

	def __init__(self, p1=None, p2=None):
		if (p1 is None and p2 is None):
			p1 = _DEFAULT_LINE3D_POINT1
			p2 = _DEFAULT_LINE3D_POINT2
		self.set(p1, p2)

	def __str__(self, title=None):
		return self.getPrintStr(title)

	def __rmatmul__(self, first):
		return self.getMultiplied(first)

	if (ENABLE_MULOPERATOR_FOR_LINE3D_CLASS):
		def __rmul__(self, first):
			return self.__rmatmul__(first)

	@classmethod
	def getClassName(cls):
		return line3d._classname

	@classmethod
	@property
	def classname(cls):
		return cls.getClassName()

	def getArray(self):
		return self._pt

	@property
	def array(self):
		return self.getArray()

	@array.setter
	def array(self, second):
		self.set(second)

	def get(self):
		return self.array()

	def getPoint(self, index=None):
		if (index is None):
			return getPoints()
		else:
			return self._pt[index]

	@property
	def point0(self):
		return self.getPoint(0)

	@property
	def point1(self):
		return self.getPoint(1)

	@property
	def p0(self):
		return self.getPoint(0)

	@property
	def p1(self):
		return self.getPoint(1)

	def getPoints(self):
		return self.array()

	@property
	def points(self):
		return self.getPoints()

	def getOrigin(self):
		return self._pt[0]

	@property
	def origin(self):
		return self.getOrigin()

	@property
	def orig(self):
		return self.getOrigin()

	def getVector(self):
		return self._pt[1] - self._pt[0]

	@property
	def vector(self):
		return self.getVector()

	@property
	def vec(self):
		return self.getVector()

	@property
	def v(self):
		return self.getVector()

	def getLength(self):
		return np.linalg.norm(self.getVector())

	@property
	def length(self):
		return self.getLength()

	@property
	def len(self):
		return self.getLength()

	@property
	def l(self):
		return self.getLength()

	def getDirection(self):
		return vecnormalize(self.getVector())

	@property
	def direction(self):
		return self.getDirection()

	@property
	def direct(self):
		return self.getDirection()

	def set(self, p1, p2=None):
		if (p2 is None):
			if (type(p1) is list or type(p1) is tuple):
				if (len(p1) == 2):
					p2 = p1[1]
					p1 = p1[0]
				else:
				 	___ERROR___
			else:
			 	___ERROR___
		if (type(p1) is np.ndarray):
			pass
		elif (type(p1) is list or type(p1) is tuple):
			p1 = np.array(p1)
		else:
			___ERROR___
		if (type(p2) is np.ndarray):
			pass
		elif (type(p2) is list or type(p2) is tuple):
			p2 = np.array(p2)
		else:
			___ERROR___
		self._pt = [p1, p2]

	def setPoints(self, pts):
		self._pt = pts

	@points.setter
	def points(self, pts):
		self.setPoints(pts)

	def setPoint(self, index, pt):
		self._pt[index] = pt

	def setPoint0(self, p):
		self.setPoint(0, pt)

	def setPoint1(self, p):
		self.setPoint(1, pt)

	@point0.setter
	def point0(self, pt):
		self.setPoint0(pt)

	@point1.setter
	def point1(self, pt):
		self.setPoint1(pt)

	@p0.setter
	def p0(self, pt):
		self.setPoint0(pt)

	@p1.setter
	def p1(self, pt):
		self.setPoint1(pt)

	def setOrigin(self, p):
		self.setPoint0(p)

	@origin.setter
	def origin(self, p):
		self.setOrigin(p)

	@orig.setter
	def orig(self, p):
		self.setOrigin(p)

	def setVector(self, v):
		self._pt[1] = self._pt[0] + v

	@vector.setter
	def vector(self, v):
		self.setVector(v)

	@vec.setter
	def vec(self, v):
		self.setVector(v)

	@v.setter
	def v(self, vec):
		self.setVector(v)

	def setLength(self, len):
		self.setVector(len * self.getDirection())

	@length.setter
	def length(self, len):
		self.setLength(len)

	@len.setter
	def len(self, len):
		self.setLength(len)

	@l.setter
	def l(self, len):
		self.setLength(len)

	def getMultiplied(self, r):
		if (type(r) is csrot3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is cstrans3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is list or type(r) is tuple):
			return self.getMultiplied(np.array(r))
		elif (type(r) is np.ndarray):
			if (r.ndim == 2):
				if (r.shape == (3, 3)):
					p0 = r @ self._pt[0]
					p1 = r @ self._pt[1]
					return line3d(p0, p1)
				elif (r.shape == (4, 4)):
					p0 = vec3(r @ vec4(self._pt[0]))
					p1 = vec3(r @ vec4(self._pt[1]))
					return line3d(p0, p1)
				else:
					___ERROR___
			else:
				___ERROR___
		else:
			___ERROR___

	def multiplied(self, r):
		return self.getMultiplied(r)

	def getRotated(self, r):
		if (type(r) is csrot3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is cstrans3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is list or type(r) is tuple):
			return self.getMultiplied(np.array(r))
		elif (type(r) is np.ndarray):
			if (r.shape == (3, 3)):
				p0 = r @ self._p[0]
				p1 = r @ self._p[1]
				return line3d(p0, p1)
			elif (r.shape == (4, 4)):
				r = to3x3(r)
				p0 = r @ self._p[0]
				p1 = r @ self._p[1]
				return line3d(p0, p1)
			else:
				___ERROR___
		else:
			___ERROR___

	def rotated(self, r):
		return getRotated(r)

	def getTranslated(self, t):
		if (type(t) is cstrans3d):
			return self.getTranslated(t.getTranslation())
		elif (type(t) is list or type(t) is tuple):
			return self.getTranslated(np.array(t))
		elif (type(t) is np.ndarray):
			p0 = self._p[0] + t
			p1 = self._p[1] + t
			return line3d(p0, p1)
		else:
			___ERROR___

	def translated(self, t):
			return self.getTranslated(t)

	def getTransformed(self, t):
		return self.getMultiplied(t)

	def transformed(self, t):
		return self.getTransformed(t)

	def getPrintStr(self, title=None):
		if (title is None):
			mesg = ''
		else:
			mesg = title + ': '
		p0 = self._pt[0]
		p1 = self._pt[1]
		mesg += "({0: .3f}, {1: .3f}, {2: .3f}) - ({3: .3f}, {4: .3f}, {5: .3f})".format(p0[0], p0[1], p0[2], p1[0], p1[1], p1[2])
		return mesg

	@property
	def printstr(self):
		return self.getPrintStr()

	@property
	def str(self):
		return self.getPrintStr()

	def print(self, title=None):
		print(self.getPrintStr(title))

	def getDataStr(self):
		return '{0:21.12f}, {1:21.12f}, {2:21.12f}, {3:21.12f}, {4:21.12f}, {5:21.12f}\n'.format(
		self._pt[0][0], self._pt[0][1], self._pt[0][2],
		self._pt[1][0], self._pt[1][1], self._pt[1][2])

	@property
	def datastr(self):
		return self.getDataStr()

	def setDataStr(self, s):
		s = s.split(',')
		for i in range(len(s)):
			s[i] = float(s[i].strip())
		self.set([(s[0], s[1], s[2]), (s[3], s[4], s[5])])

	@datastr.setter
	def datastr(self, s):
		return setDataStr(s)

	def load(self, filename):
		with open(filename, 'r') as f:
			self.setDataStr(f.read())

	def save(self, filename):
		with open(filename, 'w') as f:
			f.write(self.getDataStr())

_DEFAULT_PLANE3D_ORIGIN = [0.0, 0.0, 0.0]
_DEFAULT_PLANE3D_NORMAL = [0.0, 0.0, 1.0]

class plane3d:
	_classname = 'nkj.math.plane3d'

	def __init__(self, origin=None, normal=None):
		if (origin is None and normal is None):
			origin = _DEFAULT_PLANE3D_ORIGIN
			normal = _DEFAULT_PLANE3D_NORMAL
		self.set(origin, normal)
	
	def __str__(self, title=None):
		return self.getPrintStr(title)

	@classmethod
	def getClassName(cls):
		return plane3d._classname

	@classmethod
	@property
	def classname(cls):
		return cls.getClassName()

	def getOrigin(self):
		return self._origin

	@property
	def origin(self):
		return self.getOrigin()

	@property
	def orig(self):
		return self.getOrigin()

	def getNormal(self):
		return self._normal

	@property
	def normal(self):
		return self.getNormal()

	@property
	def n(self):
		return self.getNormal()

	def getVector(self):
		return self.getNormal()

	@property
	def vector(self):
		return self.getNormal()

	@property
	def v(self):
		return self.getNormal()

	def set(self, origin, normal=None):
		if (normal is None):
			if (type(origin) is list or type(origin) is tuple):
				if (len(origin) == 2):
					normal = origin[1]
					origin = origin[0]
				else:
				 	___ERROR___
			else:
			 	___ERROR___
		if (type(origin) is np.ndarray):
			pass
		elif (type(origin) is list or type(origin) is tuple):
			origin = np.array(origin)
		else:
			___ERROR___
		if (type(normal) is np.ndarray):
			pass
		elif (type(normal) is list or type(normal) is tuple):
			normal = np.array(normal)
		else:
			___ERROR___
		self._origin = origin
		self._normal = vecnormalize(normal)

	def setOrigin(self, origin):
		self._origin = origin

	@origin.setter
	def origin(self, origin):
		self.setOrigin(origin)

	@orig.setter
	def orig(self, origin):
		self.setOrigin(origin)

	def setNormal(self, normal):
		self._normal = normal

	@normal.setter
	def normal(self, n):
		self.setNormal(n)

	@n.setter
	def n(self, normal):
		self.setNormal(normal)

	def getMultiplied(self, r):
		if (type(r) is csrot3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is cstrans3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is list or type(r) is tuple):
			return self.getMultiplied(np.array(r))
		elif (type(r) is np.ndarray):
			if (r.ndim == 2):
				if (r.shape == (3, 3)):
					origin = r @ self._origin
					normal = r @ self._normal
					return plane3d(origin, normal)
				elif (r.shape == (4, 4)):
					origin = vec3(r @ vec4(self._origin))
					normal = to3x3(r) @ self._normal
					return plane3d(origin, normal)
				else:
					___ERROR___
			else:
				___ERROR___
		else:
			___ERROR___

	def multiplied(self, r):
		return self.getMultiplied(r)

	def getRotated(self, r):
		if (type(r) is csrot3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is cstrans3d):
			return self.getMultiplied(r.getMatrix())
		elif (type(r) is list or type(r) is tuple):
			return self.getMultiplied(np.array(r))
		elif (type(r) is np.ndarray):
			if (r.shape == (3, 3)):
				normal = r @ self._normal
				return plane3d(self._origin, normal)
			elif (r.shape == (4, 4)):
				m = to3x3(r)
				normal = m @ self._normal
				return plane3d(self._origin, normal)
			else:
				___ERROR___
		else:
			___ERROR___

	def rotated(self, r):
		return getRotated(r)

	def getTranslated(self, t):
		if (type(t) is cstrans3d):
			return self.getTranslated(t.getTranslation())
		elif (type(t) is list or type(t) is tuple):
			return self.getTranslated(np.array(t))
		elif (type(t) is np.ndarray):
			origin = self._origin + t
			return plane3d(origin, self._normal)
		else:
			___ERROR___

	def translated(self, t):
			return self.getTranslated(t)

	def getTransformed(self, t):
		return self.getMultiplied(t)

	def transformed(self, t):
		return self.getTransformed(t)

	def getPrintStr(self, title=None):
		if (title is None):
			mesg = ''
		else:
			mesg = title + ': '
		mesg += "Origin: ({0: .3f}, {1: .3f}, {2: .3f}), Normal: ({3: .3f}, {4: .3f}, {5: .3f})".format(self._origin[0], self._origin[1], self._origin[2], self._normal[0], self._normal[1], self._normal[2])
		return mesg

	@property
	def printstr(self):
		return self.getPrintStr()

	@property
	def str(self):
		return self.getPrintStr()

	def print(self, title=None):
		print(self.getPrintStr(title))

	def getDataStr(self):
		return '{0:21.12f}, {1:21.12f}, {2:21.12f}, {3:21.12f}, {4:21.12f}, {5:21.12f}\n'.format(
		self._origin[0], self._origin[1], self._origin[2], self._normal[0], self._normal[1], self._normal[2])

	@property
	def datastr(self):
		return self.getDataStr()

	def setDataStr(self, s):
		s = s.split(',')
		for i in range(len(s)):
			s[i] = float(s[i].strip())
		self.set([(s[0], s[1], s[2]), (s[3], s[4], s[5])])

	@datastr.setter
	def datastr(self, s):
		return setDataStr(s)

	def load(self, filename):
		with open(filename, 'r') as f:
			self.setDataStr(f.read())

	def save(self, filename):
		with open(filename, 'w') as f:
			f.write(self.getDataStr())

class nkjlist(list):
	_classname = 'nkj.math.nkjlist'

	def __init__(self, ls=None):
		super().__init__()
		self.set(ls)

	def __str__(self, title=None):
		return self.getPrintStr(title)

	@classmethod
	def getClassName(cls):
		return line3dlist._classname

	def get(self, index):
		return self[index]

	def set(self, ls=None):
		if (ls is None):
			super().clear()
		else:
			super().clear()
			super().extend(ls)   # 他のリストの要素を追加

	def getListLength(self):
		return len(self)

	def getLength(self):
		return self.getListLength()

	@property
	def listlength(self):
		return self.getListLength()

	@property
	def length(self):
		return self.getListLength()

	@property
	def len(self):
		return self.getListLength()

	@property
	def l(self):
		return self.getListLength()

	def getArray(self):
		outarray = []
		for i in range(len(self)):
			outarray.append(self[i].getArray())
		return np.array(outarray, dtype=np.float32)

	@property
	def array(self):
		return self.getArray()

	def getList(self):
		outlist = []
		for i in range(len(self)):
			outlist.append(self[i].getList())
		return outlist

	@property
	def list(self):
		return self.getList()

	def getPrintStr(self, title=None):
		listlen = self.getListLength()
		if (title is None):
			mesg = '--- [{0}] ---\n'.format(listlen)
		else:
			mesg = '--- {0}[{1}] ---\n'.format(title, listlen)
		for i in range(listlen):
			mesg += '{0}\n'.format(self[i])
		mesg += '---'
		return mesg

	@property
	def printstr(self):
		return self.getPrintStr()

	@property
	def str(self):
		return self.getPrintStr()

	def print(self, title=None):
		print(self.getPrintStr(title))

	def load(self, filename, datatype=None):
		self.clear()
		if (datatype is None):
			pass
		elif (datatype == 'point3d'):
			x = point3d()
		elif (datatype == 'line3d'):
			x = line3d()
		elif (datatype == 'plane3d'):
			x = plane3d()
		else:
			___NOT_SUPPORTED___
		if (datatype is None):
			with open(filename, 'r') as f:
				s = f.read()
				s = s.strip()
				s = s.split('\n')
				for linestr in s:
					self.append(linestr)
		else:
			with open(filename, 'r') as f:
				s = f.read()
				s = s.strip()
				s = s.split('\n')
				for linestr in s:
					x.setDataStr(linestr)
					self.append(copy.deepcopy(x))

	def save(self, filename):
		with open(filename, 'w') as f:
			for i in range(self.getListLength()):
				x = self[i]
				if (is_geometric(x)):
					f.write(x.getDataStr())
				else:
					f.write('{0}\n'.format(x))

class point3dlist(nkjlist):
	_classname = 'nkj.math.point3dlist'

	def __init__(self, ls=None):
		super().__init__()
		if (ls is not None):
			self.set(ls)

	def __str__(self, title=None):
		return self.getPrintStr(title)

	@classmethod
	def getClassName(cls):
		return line3dlist._classname

	def set(self, ls=None):
		super().clear()
		ty = type(ls)
		if (ls is None):
			super().clear()
		elif (ty is list):
			self.extend(ls)
		else: # リストでなく、要素が入力された
			self.append(ls)
	
	def append(self, x):
		ty = type(x)
		if (ty is point3d):
			super().append(x)
		elif (ty is list or ty is tuple or ty is np.ndarray):
			if (shape(x) == (3, )):
				super().append(point3d(x))
			else:
				___ERROR___
		else:
			___ERROR___

	def extend(self, ls):
		ty = type(ls)
		if (ty is list or ty is point3dlist):
			for x in ls:
				self.append(x)
		else:
			___ERROR___

	def load(self, filename):
		super().load(filename, 'point3d')

	def save(self, filename):
		super().save(filename, 'point3d')

class line3dlist(nkjlist):
	_classname = 'nkj.math.line3dlist'

	def __init__(self, ls=None):
		super().__init__()
		if (ls is not None):
			self.set(ls)

	def __str__(self, title=None):
		return self.getPrintStr(title)

	@classmethod
	def getClassName(cls):
		return line3dlist._classname

	def set(self, ls=None):
		super().clear()
		ty = type(ls)
		if (ls is None):
			super().clear()
		elif (ty is list):
			self.extend(ls)
		else: # リストでなく、要素が入力された
			self.append(ls)
	
	def append(self, x):
		if (type(x) is line3d):
			super().append(x)
		else:
			___ERROR___

	def extend(self, ls):
		ty = type(ls)
		if (ty is list or ty is line3dlist):
			for x in ls:
				self.append(x)
		else:
			___ERROR___

	def load(self, filename):
		super().load(filename, 'line3d')

class plane3dlist(nkjlist):
	_classname = 'nkj.math.plane3dlist'

	def __init__(self, ls=None):
		super().__init__()
		if (ls is not None):
			self.set(ls)

	def __str__(self, title=None):
		return self.getPrintStr(title)

	@classmethod
	def getClassName(cls):
		return plane3dlist._classname

	def set(self, ls=None):
		super().clear()
		ty = type(ls)
		if (ls is None):
			super().clear()
		elif (ty is list):
			self.extend(ls)
		else: # リストでなく、要素が入力された
			self.append(ls)
	
	def append(self, x):
		if (type(x) is plane3d):
			super().append(x)
		else:
			___ERROR___

	def extend(self, ls):
		ty = type(ls)
		if (ty is list or ty is plane3dlist):
			for x in ls:
				self.append(x)
		else:
			___ERROR___

	def load(self, filename):
		super().load(filename, 'plane3d')

#-- main

if __name__ == '__main__':

	ns.lib_debuglevel(_LIB_DEBUGLEVEL)

	if (True):
		print("===== complex =====")

		c = complex(1, 2)
		print(c.getClassName())
		print(c)

		print(c.normalized())
		print(normalized(c))

		print(c.conjugated())
		print(conjugated(c))

		c[0] = 10
		c[1] = 20
		print(c)

		c2 = complex(1, 3)
		print(c2)
		print(c * c2)

	if (False):
		print("===== quaternion =====")

	if (False):
		print("===== csrot3d =====")

		theta = 90 # degrees
		v = np.array([0, 0, 1])

		theta = theta / 180 * np.pi # degrees to rad
		hftheta = theta / 2
		c = np.cos(hftheta)
		s = np.sin(hftheta)
		r3 = csrot3d([c, s * v[0], s * v[1], s * v[2]])

		print("r3.w:  {0}".format(r3.w))
		print("r3.x:  {0}".format(r3.x))
		print("r3.y:  {0}".format(r3.y))
		print("r3.z:  {0}".format(r3.z))

		print("r3.rx: {0}".format(r3.rx))
		print("r3.ry: {0}".format(r3.ry))
		print("r3.rz: {0}".format(r3.rz))

		print("quat:    {0}".format(r3.quat))
		print("np.quat: {0}".format(r3.npquat))
		print("np.quat: {0} (array)".format(r3.npquatarray))
		print("R.quat:  {0}".format(r3.rquat))
		print("R.quat:  {0} (array)".format(r3.rquatarray))

		print("--- csrot3d.matrix ---")
		print(r3.matrix)
		print("---")

	if (False):
		print("===== cstrans3d =====")

		theta = 90.0 # degrees
		rv = np.array([0, 0, 1])
		tv = np.array([2, 3, 4])

		theta = theta / 180.0 * np.pi # deg 2 rad
		hftheta = theta / 2.0
		c = np.cos(hftheta)
		s = np.sin(hftheta)

		t = cstrans3d([np.array([c, s * rv[0], s * rv[1], s * rv[2]]), tv])
		t.print("cstrans3d")
		print("--- matrix ---")
		print(t.matrix)
		print("---")
		print("--- rotmat ---")
		print(t.rotmat)
		print("---")
		print("translation vector: {0}".format(t.trans))

		t_i = t.inv
		t_i.print("inversed")
		print("--- inversed matrix ---")
		print(t_i.matrix)
		print("---")

		t2 = t @ t_i
		print("--- matrix ---")
		print(t2.matrix)
		print("---")
		mat4x4(t2).print('test matrix')

		t3 = t_i @ t
		print("--- matrix ---")
		print(t3.matrix)
		print("---")
		mat4x4(t3).print('test matrix')

	if (False):
		print("===== matrix =====")

		a = np.array([[1, 0], [0, 2]])
		print(a)
		print(type(a))

		m = mat4x4()
		print("class name: ", str_bracket(m.getClassName()))
		print("type:       ", type(m))
		print("dimension:  ", m.dim())
		print("size:       ", m.size())
		m.print("m")
		print("--- m.array() ---")
		print(m.array())
		print("---")

		m.array()[0][1] = 3.0
		print(m)

		m2 = mat4x4()

		print(m @ m2)

		m3 = m @ m2
		print(m3)

		m3.getInversed().print('inversed')

		m3.print('original')

		m4 = mat4x4(m3.get())
		m4.print('m4')

		"""
		m5 = mat4x4(np.array([deg2rad(30), 0, 0, 1]))
		m5.print('m5')
		"""

		(~m3).print('invert')

		(~m3 * m3).print('validation')
		(~m3 @ m3).print('validation')

		m3.inverse()
		m3.print('inverse')

		m3.identity()
		m3.print('identity')

	if (False):
		print("===== sigmoid =====")
		import matplotlib.pyplot as plt

		x = np.arange(-10, 10, 0.1)
		ta = 1.0
		x0 = 2.5
		#y = sigmoid(x, tau=tau, x0=x0)
		y = sigmoid(x, x0=x0)
		plt.plot(x, y)
		plt.show()

	if (False):
		print("===== log ticks =====")

		ticks = logticks(0.01, 2.0)
		print(ticks)

	if (False):
		print("===== vec3d =====")

		print(vec3d())

		v = vec3d(np.array([1, 0, 0]))
		print(v)

		v0 = vec3d(np.array([0, 1, 0]))

		print("angle: {0} degrees".format(rad2deg(angle(v0, v))))

	if (True):
		print("===== point3d =====")

		print(point3d())

		p = point3d(np.array([1, 1, 2]))
		print(p)

	if (False):
		print("===== line3d =====")

		print(line3d())

		l = line3d(np.array([0, 0, 0]), np.array([0, 0, 100]))
		l.print('line3d')
		print('origin: {0}'.format(l.getOrigin()))
		print('length: {0}'.format(l.getLength()))
		print('vector: {0}'.format(l.getVector()))
		print('direction: {0}'.format(l.getDirection()))

		l2 = line3d([[1, 1, 1], [2, 2, 2]])
		l2.print('line3d 2')

		theta = deg2rad(90)
		t = cstrans3d()
		ver = 1
		if (ver == 1):
			t.rotateGlobal([deg2rad(45), (0, 1, 0)], 'angleaxis')
			t.rotateGlobal([deg2rad(theta), (0, 0, 1)], 'angleaxis')
		elif (ver == 2):
			t.rotateGlobal([deg2rad(45), (np.cos(theta), np.sin(theta), 0)], 'angleaxis')
		else:
			___ERROR___
		t.print('transformation')
		#l2 = l.transformed(t)
		l2 = t @ l
		l2.print('line3d (transformed)')

		l2.save('testdata/test.line')
		l2.load('testdata/test.line')
		l2.print('reload')

		#-- direction cosine test

		xlist = []
		zlist = []
		philist = []
		thetalist = []
		for zdeg in range(-90, 91, 5):
			ztheta = deg2rad(zdeg)
			for xdeg in range(-90, 91, 5):
				xtheta = deg2rad(xdeg)
				dprint('ztheta:{0:.3f}, xtheta:{1:.3f}'.format(rad2deg(ztheta), rad2deg(xtheta)))
				phi = np.sqrt(ztheta**2 + xtheta**2)
				if (phi > np.pi / 2):
					phi = np.pi - phi
				dprint('phi:{0:.3f}'.format(rad2deg(phi)))
				c_phi = np.cos(phi)
				s_phi = np.sin(phi)
				theta = np.arctan2(ztheta, xtheta)
				c = s_phi * np.cos(theta)
				s = s_phi * np.sin(theta)
				dprint([c, c_phi, s])
				r = csrot3d([c, c_phi, s], 'dircos', [0, 1, 0])
				#r.print('direction cosine')
				xlist.append(xdeg)
				zlist.append(zdeg)
				philist.append(rad2deg(phi))
				thetalist.append(theta)
		xlist = np.array(xlist)
		zlist = np.array(zlist)
		philist = np.array(philist)
		thetalist = np.array(thetalist)

		ticks = int(np.sqrt(len(xlist)))
		xlist = xlist.reshape(ticks, ticks)
		zlist = zlist.reshape(ticks, ticks)
		philist = philist.reshape(ticks, ticks)
		thetalist = thetalist.reshape(ticks, ticks)
		dprint('xlist[{0}]: {1}'.format(len(xlist), xlist))
		dprint('zlist[{0}]: {1}'.format(len(zlist), zlist))
		dprint('philist[{0}]: {1}'.format(len(philist), philist))

		from matplotlib import pyplot as plt

		plt.pcolor(xlist, zlist, philist, cmap='inferno', vmin=0, vmax=90)
		""" #--- colormaps
		plt.pcolor(xlist, zlist, philist, cmap='plasma', vmin=0, vmax=90)
		plt.pcolor(xlist, zlist, philist, cmap='bwr', vmin=0, vmax=90)
		plt.pcolor(xlist, zlist, philist, cmap='coolwarm', vmin=0, vmax=90)
		plt.pcolor(xlist, zlist, philist, cmap='seismic', vmin=0, vmax=90)
		plt.pcolor(xlist, zlist, philist, cmap='RdYlBu_r', vmin=0, vmax=90)
		plt.pcolor(xlist, zlist, thetalist)
		"""
		plt.xticks([-90, -60, -30, 0, 30, 60, 90], fontsize=15)
		plt.yticks([-90, -60, -30, 0, 30, 60, 90], fontsize=15)
		plt.xlabel('Angle of x direction cosine', fontsize=18)
		plt.ylabel('Angle of z direction cosine', fontsize=18)
		plt.tight_layout()   # フォントサイズを変更したとき、図で文字が見切れないように調整する。
		colorbar_ticks = np.linspace(0, 90, 7, endpoint=True)
		if (True):
			cb = plt.colorbar(ticks=colorbar_ticks)
			for t in cb.ax.get_yticklabels():
				t.set_fontsize(12)
		else:
			plt.colorbar(ticks=colorbar_ticks)
		plt.savefig('testdata/test.png')
		plt.show()

	if (False):
		print("===== plane3d =====")

		pl = plane3d(np.array([0, 0, 0]), np.array([0, 0, 100]))
		pl.print('plane3d')

		pl2 = plane3d([np.array([0, 0, 0]), np.array([0, 10, 0])])
		pl2.print('plane3d 2')

		pl3 = plane3d([[0, 0, 0], [1, 1, 0]])
		pl3.print('plane3d 3')

		print('{0} degree'.format(rad2deg(angle(pl, pl2))))
		print('{0} degree'.format(rad2deg(angle(pl2, pl3))))

		theta = deg2rad(45)
		phi = deg2rad(45)
		print('theta:{0:.3f}, phi:{1:.3f}'.format(rad2deg(theta), rad2deg(phi)))
		t = cstrans3d()
		ver = 2
		if (ver == 1):
			t.rotateGlobal([phi, (np.cos(theta), np.sin(theta), 0)], 'angleaxis')
		elif (ver == 2):
			t.rotateGlobal([phi, (1, 0, 0)], 'angleaxis')
			t.rotateGlobal([theta, (0, 0, 1)], 'angleaxis')
		else:
			___ERROR___
		t.translateGlobal([10, 10, 0])
		t.print('transformation')
		pl2 = t @ pl
		pl2.print('plane3d (transformed)')

		pl2.save('testdata/test.plane')
		pl2.load('testdata/test.plane')
		pl2.print('reload')

	if (False):
		print("===== nkjlist =====")

		ls = nkjlist()
		ls.print('nkjlist')

		ls.append('test')
		ls.append('test2')
		ls.print('nkjlist')

		ls.set(['test3'])
		ls.print('nkjlist (test 3)')

		ls.clear()
		ls.print('nkjlist')

		ls.clear()
		ls.append([1, 2, 3])  # リストを一要素として追加
		ls.extend([3, 4, 5])  # 要素をリストで追加
		ls.save('testdata/test.list')
		ls.load('testdata/test.list')
		ls.print('nkjlist')

	"""
	if (False):
		print("===== vec3dlist =====")

		ls = vec3dlist()
		ls.append([0, 1, 2])
		ls.append([3, 4, 5])
		ls.extend([[6, 7, 8], [9, 10, 11]])
		ls.print('vec3dlist')
		ls.save('testdata/test.pts')
		ls.load('testdata/test.pts')
		ls.print('vec3dlist')
	"""

	if (False):
		print("===== point3dlist =====")

		ls = point3dlist()
		ls.append([0, 1, 2])
		ls.append([3, 4, 5])
		ls.extend([[6, 7, 8], [9, 10, 11]])
		ls.print('point3dlist')
		ls.save('testdata/test.pts')
		ls.load('testdata/test.pts')
		ls.print('point3dlist')

	if (False):
		print("===== line3dlist =====")

		ls = line3dlist()
		ls.append(line3d([0, 1, 2], [0, 1, 0]))
		ls.append(line3d([3, 4, 5], [3, 4, 0]))
		ls.print('line3dlist')
		ls.save('testdata/test.lines')
		ls.load('testdata/test.lines')
		ls.print('line3dlist')

	if (False):
		print("===== plane3dlist =====")

		ls = plane3dlist()
		ls.append(plane3d([0, 1, 2], [0, 1, 0]))
		ls.append(plane3d([3, 4, 5], [3, 4, 0]))
		ls.extend(copy(ls))
		ls.print('plane3dlist')
		ls.save('testdata/test.planes')
		ls.load('testdata/test.planes')
		ls.print('plane3dlist')



#
# [name] nkj.fa.py
# [exec] python fa.py
# [purpose] library for frequency analysis
#
# [reference] https://qiita.com/yukiB/items/59f8484e72bb0471ad47
#             https://qiita.com/MuAuan/items/8850e037babcff991b8e
#             https://qiita.com/MuAuan/items/858aab2879708668e2bb
#
# Written by Yoshikazu NAKAJIMA
#

import os
import sys
import math
import cmath
import numpy as np
import scipy as sp
import matplotlib.pyplot as plt

sys.path.append(os.path.abspath(".."))
from nkj.str import *
from nkj.math import *
from nkj.filter import *

#-- global functions

def omega(f): # frequency to omega
	return (2.0 * np.pi * f)

def freq(f):  # omega to frequency
	return f / (2.0 * np.pi)

def f2cycletime(f): # ct: cycle time
	if (f == 0.0):
		return ValueError
	else:
		return 1.0 / f

def f2ct(f):
	return f2cycletime(f)

def cycletime2f(ct):
	if (ct == 0.0):
		return ValueError
	else:
		return 1.0 / ct

def ct2f(ct):
	return cycletime2f(ct)

def phase2time(theta, f):
	omega = f2omega(f)
	if (omega == ValueError):
		return ValueError
	else:
		return theta / omega

def time2phase(t, f):
	return f2omega(f) * t

def spiffphase(theta): # reshape phase to [0.0, 2.0 * pi]
	dpi = 2.0 * np.pi
	while (theta < 0.0):
		theta = theta + dpi
	while (theta >= dpi):
		theta = theta - dpi
	return theta

def cphase(c, t=None, f=None):
	if (t == None):
		phase = cmath.phase(c)
	else:
		phase = cmath.phase(c) - time2phase(t, f)
	return spiffphase(phase)

#-- classes

class nkjfa():
	#-- class variables
	_classname = "nkjfa"

	def __init__(self):
		ldprint("Classname: ", str_bracket(nkjfa._classname))

	def __str__(self):
		return concat("Classname: ", str_bracket(nkjfa._classname))

#-- main

if __name__ == '__main__':

	print(nkjfa)

	#-- FFT

	"""
	fs = 1024
	N = 10 * fs
	nperseg = 512
	amp = 2 * np.sqrt(2)
	noise_power = 0.001 * fs / 2
	time = np.arange(N) / float(fs)
	carrier = amp * np.sin(2 * np.pi * 50 * time)

	noise = np.random.normal(scale=np.sqrt(noise_power), size=time.shape)

	x = carrier + noise

	f, t, Zxx = sp.stft(x, fs=fs, nperseg=nperseg)

	plt.figure()
	plt.pcolormesh(t, f, np.abs(Zxx), vmin=0, vmax=amp)
	plt.ylim([f[1], f[-1]])
	plt.title('STFT magnitude')
	plt.ylabel('Frequency [Hz]')
	plt.xlabel('Time [sec]')
	plt.yscale('log')
	plt.show()
	"""

	print("phase:", rad2deg(spiffphase(deg2rad(718.2))))
	print("phase:", rad2deg(spiffphase(deg2rad(-90.2))))



#
# [name] nkj.curvefit.py
# [exec] python3 -m nkj.scipy
#
# Written by Yoshikazu NAKAJIMA

import numpy as np
import scipy as sp
from scipy.optimize import curve_fit

import nkj as nk

# Gauss

def gauss_fit_mu0(x, y, mu0):
	nk.ldprint(["--> gauss_fit_mu0(,, mu0:{}".format(mu0)])
	mu = mu0
	def gauss_f(x, sigma, a):
		#
		#                 (x - mu)^2
		# f(x) = a exp(- -------------)
		#                 2 * sigma^2
		#
		return a * np.exp(-(x - mu)**2 / (2 * sigma**2))

	popt, pcov = curve_fit(gauss_f, x, y)
	nk.ldprint2(["popt: {}".format(popt)])
	nk.ldprint2(["pcov: {}".format(pcov)])

	sigma = popt[0]
	a = popt[1]

	nk.ldprint2("sigma: {}".format(sigma))
	nk.ldprint2("mu:    {}".format(mu))
	nk.ldprint2("a:     {}".format(a))

	nk.ldprint(["<-- sigmiod_fit(): sigma:{0:.6f}, a:{1:.6f}".format(sigma, a)])
	return sigma, a

def gauss_fit(x, y, mu=None):
	nk.ldprint(["--> gauss_fit(,, mu:{}".format(mu)])

	if (mu is not None):
		sigma, a = gauss_fit_mu0(x, y, mu)
		nk.ldprint(["<-- gauss_fit(): sigma:{0:.6f}, mu:{1:.6f}, a:{2:.6f}".format(sigma, mu, a)])
		return sigma, mu, a

	def gauss_f(x, sigma, mu, a):
		#
		#                 (x - mu)^2
		# f(x) = a exp(- -------------)
		#                 2 * sigma^2
		#
		return a * np.exp(-(x - mu)**2 / (2 * sigma**2))

	popt, pcov = curve_fit(gauss_f, x, y)
	nk.ldprint2(["popt: {}".format(popt)])
	nk.ldprint2(["pcov: {}".format(pcov)])

	sigma = popt[0]
	mu = popt[1]
	a = popt[2]

	nk.ldprint2("sigma: {}".format(sigma))
	nk.ldprint2("mu:    {}".format(mu))
	nk.ldprint2("a:     {}".format(a))

	nk.ldprint(["<-- gauss_fit(): sigma:{0:.6f}, mu:{1:.6f}, a:{2:.6f}".format(sigma, mu, a)])
	return sigma, mu, a

# Sigmoid

def sigmoid_fit_b0a0(x, y, b0=0.0, a0=1.0):
	nk.ldprint(["--> sigmiod_fit_b0a0(,, b0:{0}, a0:{1})".format(b0, a0)])
	b = b0
	a = a0
	def sigmoid_f_b0a0(x, tau, x0):
		#
		#                    a
		# f(x) = -------------------------- + b
		#         1 + exp(-(x - x0) / tau)
		#
		return a / (1.0 + np.exp(-(x - x0) / tau)) + b

	popt, pcov = curve_fit(sigmoid_f_b0a0, x, y)
	nk.ldprint2(["popt: {}".format(popt)])
	nk.ldprint2(["pcov: {}".format(pcov)])

	tau = popt[0]
	x0 = popt[1]

	nk.ldprint2("tau: {}".format(tau))
	nk.ldprint2("x0:  {}".format(x0))

	nk.ldprint(["<-- sigmiod_fit_b0a0(): tau:{0:.6f}, x0:{1:.6f}".format(tau, x0)])
	return tau, x0

def sigmoid_fit_a0(x, y, a0=1.0):
	nk.ldprint(["--> sigmiod_fit_a0()"])
	a = a0
	def sigmoid_f_a0(x, tau, x0, b):
		#
		#                    a
		# f(x) = -------------------------- + b
		#         1 + exp(-(x - x0) / tau)
		#
		return a / (1.0 + np.exp(-(x - x0) / tau)) + b

	popt, pcov = curve_fit(sigmoid_f_a0, x, y)
	nk.ldprint2(["popt: {}".format(popt)])
	nk.ldprint2(["pcov: {}".format(pcov)])

	tau = popt[0]
	x0 = popt[1]
	b = popt[2]

	nk.ldprint2("tau: {}".format(tau))
	nk.ldprint2("x0:  {}".format(x0))
	nk.ldprint2("b:   {}".format(b))

	nk.ldprint(["<-- sigmiod_fit_a0(): tau:{0:.6f}, x0:{1:.6f}, b:{2:.6f}".format(tau, x0, b)])
	return tau, x0, b

def sigmoid_fit_b0(x, y, b0=0.0):
	nk.ldprint(["--> sigmiod_fit_b0()"])
	b = b0
	def sigmoid_f_b0(x, tau, x0, a):
		#
		#                    a
		# f(x) = -------------------------- + b
		#         1 + exp(-(x - x0) / tau)
		#
		return a / (1.0 + np.exp(-(x - x0) / tau)) + b

	popt, pcov = curve_fit(sigmoid_f_b0, x, y)
	nk.ldprint2(["popt: {}".format(popt)])
	nk.ldprint2(["pcov: {}".format(pcov)])

	tau = popt[0]
	x0 = popt[1]
	a = popt[2]

	nk.ldprint2("tau: {}".format(tau))
	nk.ldprint2("x0:  {}".format(x0))
	nk.ldprint2("a:   {}".format(a))

	nk.ldprint(["<-- sigmiod_fit_b0(): tau:{0:.6f}, x0:{1:.6f}, a:{2:.6f}".format(tau, x0, a)])
	return tau, x0, a

def sigmoid_fit(x, y, b0=None, a0=None):
	nk.ldprint(["--> sigmoid_fit(,, b0:{0}, a0:{1})".format(b0, a0)])
	if (b0 is not None):
		if (a0 is not None):  # fix b and a
			tau, x0 = sigmoid_fit_b0a0(x, y, b0, a0)
			nk.ldprint(["<-- sigmoid_fit(): tau:{0:.6f}, x0:{1:.6f}, b:{2:.6f}, a:{3:.6f}".format(tau, x0, b0, a0)])
			return tau, x0, b0, a0
		else:  # estimate a and fix b
			tau, x0, a = sigmoid_fit_b0(x, y, b0)
			nk.ldprint(["<-- sigmoid_fit(): tau:{0:.6f}, x0:{1:.6f}, b:{2:.6f}, a:{3:.6f}".format(tau, x0, b0, a)])
			return tau, x0, b0, a
	elif (a0 is not None):  # estimate b and fix a
			tau, x0, b = sigmoid_fit_a0(x, y, b0)
			nk.ldprint(["<-- sigmoid_fit(): tau:{0:.6f}, x0:{1:.6f}, b:{2:.6f}, a:{3:.6f}".format(tau, x0, b, a0)])
			return tau, x0, b, a0

	# estimate b and a

	def sigmoid_f(x, tau, x0, b, a):
		#
		#                    a
		# f(x) = -------------------------- + b
		#         1 + exp(-(x - x0) / tau)
		#
		return a / (1.0 + np.exp(-(x - x0) / tau)) + b

	popt, pcov = curve_fit(sigmoid_f, x, y)
	nk.ldprint2(["popt: {}".format(popt)])
	nk.ldprint2(["pcov: {}".format(pcov)])

	tau = popt[0]
	x0 = popt[1]
	b = popt[2]
	a = popt[3]

	nk.ldprint2("tau: {}".format(tau))
	nk.ldprint2("x0:  {}".format(x0))
	nk.ldprint2("b:   {}".format(b))
	nk.ldprint2("a:   {}".format(a))

	nk.ldprint(["<-- sigmiod_fit(): tau:{0:.6f}, x0:{1:.6f}, b:{2:.6f}, a:{3:.6f}".format(tau, x0, b, a)])
	return tau, x0, b, a

# Extreme

def extreme_fit_x0(x, y, x0:float):
	nk.ldprint(["--> extreme_fit_x0(,, {})".format(x0)])
	nk.ldprint2(["x0: {}".format(x0)])
	nk.ldprint2(["x0.type: {}".format(type(x0))])
	xc = x0
	def extreme_f(x, w, a):
		#
		#                         x - xc      x - xc
		# f(x) = A * exp [-exp(- --------) - -------- + 1] + b
		#                           w           w
		#
		# ここでは +b （バイアス）は考えない
		#
		return a * np.exp(-np.exp(-(x - xc) / w) - (x - xc) / w + 1)

	popt, pcov = curve_fit(extreme_f, x, y)
	nk.ldprint2(["popt: {}".format(popt)])
	nk.ldprint2(["pcov: {}".format(pcov)])

	w = popt[0]
	a = popt[1]

	nk.ldprint2("w:  {}".format(w))
	nk.ldprint2("a:  {}".format(a))

	nk.ldprint(["<-- extreme_fit_x0(): w:{0:.6f}, a:{1:.6f}".format(w, a)])
	return w, a

def extreme_fit(x, y, x0=None):
	nk.ldprint(["--> extreme_fit(,, {})".format(x0)])
	nk.ldprint2(["x0: {}".format(x0)])

	if (x0 is not None):
		w, a = extreme_fit_x0(x, y, x0)
		nk.ldprint(["<-- extreme_fit(): a:{0:.6f}, xc:{1:.6f}, w:{2:.6f}".format(a, x0, w)])
		return w, x0, a

	def extreme_f(x, w, xc, a):
		#
		#                         x - xc      x - xc
		# f(x) = A * exp [-exp(- --------) - -------- + 1]
		#                           w           w
		#
		return a * np.exp(-np.exp(-(x - xc) / w) - (x - xc) / w + 1)

	popt, pcov = curve_fit(extreme_f, x, y)
	nk.ldprint2(["popt: {}".format(popt)])
	nk.ldprint2(["pcov: {}".format(pcov)])

	w = popt[0]
	xc = popt[1]
	a = popt[2]

	nk.ldprint2("w:  {}".format(w))
	nk.ldprint2("xc: {}".format(xc))
	nk.ldprint2("a:  {}".format(a))

	nk.ldprint(["<-- extreme_fit(): w:{0:.6f}, xc:{1:.6f}, a:{2:.6f}".format(w, xc, a)])
	return w, xc, a




#
# [name]    nkj.ea.py
# [comment] library for error analysis
#
# Written by Yoshikazu NAKAJIMA
#

import os
import sys
import numpy as np

sys.path.append(os.path.abspath(".."))
from nkj.str import *
import nkj.math as nm
import nkj.cam as nc

def _correct_convergence_angle(a):
	ldprint(['convergence angle: {0:.3f} degrees'.format(nm.rad2deg(a))])
	if (a < 0.0):
		a = np.abs(a)
		ldprint(['convergence angle: {0:.3f} degrees (corrected)'.format(nm.rad2deg(a))])
	if (a > np.pi / 2.0):
		a = np.pi - a
		ldprint(['convergence angle: {0:.3f} degrees (corrected)'.format(nm.rad2deg(a))])
	return a

def conv_angle_point(pt, cam1, cam2):
	sight1 = cam1.line_of_sight(pt)
	sight2 = cam2.line_of_sight(pt)
	return _correct_convergence_angle(nm.angle(sight1, sight2))

def conv_angle_line(line, cam1, cam2):
	lorig = line.origin
	lv = line.direction
	cv1 = nm.vecnormalize(lorig - cam1.focalpoint)
	xaxis = np.cross(cv1, lv)
	sight1 = np.cross(lv, xaxis)
	cv2 = nm.vecnormalize(lorig - cam2.focalpoint)
	xaxis = np.cross(cv2, lv)
	sight2 = np.cross(lv, xaxis)
	return _correct_convergence_angle(nm.angle(sight1, sight2))

#--- main

if __name__ == '__main__':

	CAMERA_POS = [0.0, 0.0, -1800.0]
	CAMERA_ANGLE = -90.0   # degree
	ROTATION_AXIS = [0.0, 1.0, 0.0]
	POINT = [100.0, 0.0, 0.0]
	LINE_PT0 = [0.0, 0.0, 0.0]
	LINE_PT1 = [0.0, -100.0, 0.0]

	camangle = nm.deg2rad(CAMERA_ANGLE)
	print('camera angle: {0:.1f} degrees'.format(nm.rad2deg(camangle)))
	print('')

	cam1 = nc.nkjcam()
	cam1.translateLocal(CAMERA_POS)
	cam1.print('camera 1')
	print('')

	cam2 = nc.nkjcam()
	cam2.translateLocal(CAMERA_POS)
	cam2.rotateGlobal(camangle * np.array(ROTATION_AXIS), 'rotvec')
	cam2.print('camera 2')
	print('')

	pt = POINT
	print('point: {0}'.format(pt))
	print('')

	line = nm.line3d(LINE_PT0, LINE_PT1)
	line.print('line')
	print('')

	# compute convergence angle

	convergence_p = conv_angle_point(pt, cam1, cam2)
	convergence_l = conv_angle_line(line, cam1, cam2)

	# show result

	print('--- result ---')
	print('camera angle:                    {0:.3f} degrees'.format(nm.rad2deg(_correct_convergence_angle(camangle))))
	print('convergence angle for the point: {0:.3f} degrees'.format(nm.rad2deg(convergence_p)))
	print('convergence angle for the line:  {0:.3f} degrees'.format(nm.rad2deg(convergence_l)))
	print('---')



#
# [name] nkj.plot.py
#
# Written by Yoshikazu NAKAJIMA (Wed Apr 14 18:59:31 JST 2021)
#
_LIB_DEBUGLEVEL = 1

def barplot_annotate_brackets(num1, num2, data, center, height, yerr=None, dh=.05, barh=.05, fs=None, maxasterix=None):
    #
    # reference (Japanese): https://omedstu.jimdofree.com/2019/02/11/matplotlib%E3%81%A7%E6%A3%92%E3%82%B0%E3%83%A9%E3%83%95%E9%96%93%E3%81%AE%E6%9C%89%E6%84%8F%E5%B7%AE%E3%81%AE%E6%8F%8F%E7%94%BB%E3%82%92%E3%81%99%E3%82%8B/
    # reference (English):  https://matplotlib.org/2.0.2/examples/api/barchart_demo.html
    #
    """
    Annotate barplot with p-values.
    :param num1: number of left bar to put bracket over
    :param num2: number of right bar to put bracket over
    :param data: string to write or number for generating asterixes
    :param center: centers of all bars (like plt.bar() input)
    :param height: heights of all bars (like plt.bar() input)
    :param yerr: yerrs of all bars (like plt.bar() input)
    :param dh: height offset over bar / bar + yerr in axes coordinates (0 to 1)
    :param barh: bar height in axes coordinates (0 to 1)
    :param fs: font size
    :param maxasterix: maximum number of asterixes to write (for very small p-values)
    """

    if type(data) is str:
        text = data
    else:
        # * is p < 0.05
        # ** is p < 0.005
        # *** is p < 0.0005
        # etc.
        text = ''
        p = .05

        while data < p:
            text += '*'
            p /= 10.

            if maxasterix and len(text) == maxasterix:
                break

        if len(text) == 0:
            text = 'n. s.'

    lx, ly = center[num1], height[num1]
    rx, ry = center[num2], height[num2]

    if yerr:
        ly += yerr[num1]
        ry += yerr[num2]

    ax_y0, ax_y1 = plt.gca().get_ylim()
    dh *= (ax_y1 - ax_y0)
    barh *= (ax_y1 - ax_y0)

    y = max(ly, ry) + dh

    barx = [lx, lx, rx, rx]
    bary = [y, y+barh, y+barh, y]
    mid = ((lx+rx)/2, y+barh)

    plt.plot(barx, bary, c='black')

    kwargs = dict(ha='center', va='bottom')
    if fs is not None:
        kwargs['fontsize'] = fs

    plt.text(*mid, text, **kwargs)

#-- main



#
# [name]  time.py
# [class] nkj.time
# [exec]  py -m nkj.time
#
# Written by Yoshikazu NAKAJIMA
#

from typing import Union
from nkj.str import *
import time as pytime
import datetime

_DEFAULT_TIMEPERIOD = 0.0

class time():

	def __init__(self, val=None, tp=None):
		ldprint('--> time.__init__({0}, {1})'.format(val, tp))
		if (val is None):
			dt = None
		elif (type(val) == tuple or type(val) == list):
			dt = val[0]
			if (tp is None):  # tp の指定がないときのみ、val に指定された time period を反映（val の第2引数よりも tp の入力を優先）
				tp = val[1]
		elif (type(val) == time):
			dt = val.datetime
			if (tp is None):  # tp の指定がないときのみ、val (time 形式) に指定された time period を反映（第1引数 time クラス内の timeperiod よりも tp の入力を優先）
				tp = val.timeperiod
		else:
			dt = val

		if (dt is None):
			self._datetime = datetime.datetime.now()
		elif (type(dt) == datetime.datetime):
			self._datetime = dt
		elif (type(dt) == float):
			self._datetime = datetime.datetime.fromtimestamp(dt)  # UNIX time
		else:
			__ERROR__NOTIMPLEMENTED

		if (self._datetime is None):
			ldprint('WARNING: NULL datetime')
			return

		ldprint2('datetime:   {}'.format(self._datetime))
		ldprint2('timeperiod: {0} ({1})'.format(tp, type(tp)))

		if (tp is None):
			self._timeperiod = _DEFAULT_TIMEPERIOD
		elif (type(tp) == float):
			self._timeperiod = tp
		elif (type(timeperiod) == int):
			self._timeperiod = float(tp)
		elif (type(timeperiod) == time):
			self._timeperiod = tp.timestamp - self.timestamp  # end

		ldprint('datetime:   {}'.format(self._datetime))
		ldprint('timeperiod: {}'.format(self._timeperiod))

		ldprint('<-- time.__init__()')

	def __add__(self, second):
		ldprint('--> time.__add()')
		ldprint('self:   {0} ({1})'.format(self.get(), type(self.get())))
		ty = type(second)
		ldprint2('second.type: {}'.format(type(second)))
		if (ty == float):
			ldprint('second: {0} ({1})'.format(second, type(second)))
			ldprint('<-- time.__add__()')
			return time(self.timestamp + second)
		elif (ty == int):
			ldprint('second: {0} ({1})'.format(second, type(second)))
			ldprint('<-- time.__add__()')
			return time(self.timestamp + float(second))
		else:
			ldprint('<-- time.__add__(): None (ERROR)')
			return None

	def __sub__(self, second):
		ldprint('--> time.__sub__()')
		ldprint('self:   {0} ({1})'.format(self.get(), type(self.get())))
		ty = type(second)
		ldprint2('second.type: {}'.format(type(second)))
		if (ty == time):
			ldprint('second: {0} ({1})'.format(second.get(), type(second)))
			ldprint('<-- time.__sub__()')
			return self.timestamp - second.timestamp  # return float
		elif (ty == datetime.datetime):
			ldprint('second: {0} ({1})'.format(second, type(second)))
			ldprint('<-- time.__sub__()')
			return self.timestamp - second.timestamp()  # return float
		elif (ty == float):
			ldprint('<-- time.__sub__()')
			return datetime.datetime.fromtimestamp(self.timestamp - second)  # return datetime
		elif (ty == int):
			ldprint('<-- time.__sub__()')
			return datetime.datetime.fromtimestamp(self.timestamp - float(second))  # return datetime
		else:
			ldprint('<-- time.__sub__(): None (ERROR)')
			return None

	def set(self, val):
		ldprint('--> time.set({0}, {1})'.format(val, type(val)))
		ty = type(val)
		if (ty == datetime.datetime):
			self._datetime = val
		elif (ty == float):
			self._datetime = datetime.datetime.fromtimestamp(val)
		else:
			self._datetime = None  # __ERROR__UNKNOWNTYPE
		ldprint('<-- time.set()')

	def get(self):
		return self._datetime

	@property
	def timeperiod(self):
		return self._timeperiod

	@timeperiod.setter
	def timeperiod(self, tp):
		self._timeperiod = tp

	@property
	def start(self):
		return self

	@start.setter
	def start(self, val):
		self.set(val)

	@property
	def end(self):
		ldprint('self.start:      {}'.format(self.start.datetime))
		ldprint('self.datetime:   {}'.format(self.datetime))
		ldprint('self.timestamp:  {}'.format(self.timestamp))
		ldprint('self.timeperiod: {}'.format(self.timeperiod))
		ldprint(self.timestamp + self.timeperiod)
		ldprint(self + self.timeperiod)
		return self.start + self.timeperiod

	@end.setter
	def end(self, val):
		self.set(val - self.start)

	def update(self):
		self.set(datetime.datetime.now())
		return self.get()

	def refresh(self):
		return self.update()

	def now(self):
		return self.update()

	@property
	def datetime(self):
		return self.get()

	@datetime.setter
	def datetime(self, val:datetime):
		self.set(val)

	@property
	def timestamp(self):
		return self.get().timestamp()

	@timestamp.setter
	def timestamp(self, val):
		self.set(val)

	def includes(self, t):
		ty = type(t)
		ldprint2('type: {}'.format(ty))
		if (ty == time):
			ldprint2('second: {0} ({1})'.format(t.datetime, type(t)))
			ts_start = t.timestamp
			ts_end = t.end.timestamp
		elif (ty == datetime.datetime):
			ts_start = t.timestamp()
			ts_end = ts_start  # time period should be 0
		else:  # UNIX time
			ts_start = t
			ts_end = ts_start  # time period should be 0
		ldprint2('start: {}'.format(ts_start))
		ldprint2('end:   {}'.format(ts_end))
		td_start = ts_start - self.timestamp
		td_end = ts_end - self.end.timestamp
		return True if (td_start >= 0.0 and td_end <= 0.0) else False

	def included(self, t):
		ty = type(t)
		ldprint2('type: {}'.format(ty))
		if (ty == time):
			ts_start = t.timestamp
			ts_end = t.end.timestamp
		elif (ty == datetime.datetime):
			ts_start = t.timestamp()
			ts_end = ts_start  # time period should be 0
		else:  # UNIX time
			ts_start = t
			ts_end = ts_start  # time period should be 0
		td_start = ts_start - self.timestamp
		td_end = ts_end - self.end.timestamp
		return True if (td_start <= 0.0 and td_end >= 0.0) else False

#-- main

if (__name__ == '__main__'):

	_DEBUGLEVEL = 0
	_LIB_DEBUGLEVEL = 0
	debuglevel(_DEBUGLEVEL)
	lib_debuglevel(_LIB_DEBUGLEVEL)

	t = time()
	print('get():      {}'.format(t.get()))
	print('now():      {} (updated)'.format(t.now()))
	print('get():      {} (not updated)'.format(t.get()))
	print('datetime:   {} (not updated)'.format(t.datetime))
	print('timestamp:  {} (not updated)'.format(t.timestamp))

	print('\n--')
	print('t:       {}'.format(t.datetime))
	print('t + 3:   {}'.format(t + 3))
	print('t + 0.1: {}'.format(t + 0.1))
	print('t - 3:   {}'.format(t - 3))
	print('t - 0.1: {}'.format(t - 0.1))

	print('\n--')
	t.timeperiod = 3.0
	print('t:            {}'.format(t.datetime))
	print('t.start:      {}'.format(t.start.datetime))
	print('t.timeperiod: {}'.format(t.timeperiod))
	print('t.end:        {}'.format(t.end))

	print('\n--')

	def func():
		pytime.sleep(0.1)

	t1 = time()
	print('Start time: {0} ({1})'.format(t1.datetime, type(t1)))
	func()
	t2 = time()
	print('End time:   {0} ({1})'.format(t2.datetime, type(t2)))
	print('computation time: {}'.format(t2 - t1))

	print('\n--')
	print('includes: {}'.format(t1.includes(t1)))
	print('includes: {}'.format(t1.includes(t2)))
	print('includes: {}'.format(t2.includes(t1)))
	print('included: {}'.format(t1.included(t1)))
	print('included: {}'.format(t1.included(t2)))
	print('included: {}'.format(t2.included(t1)))

	print('\n--')
	t1.timeperiod = 1.0
	print('includes: {}'.format(t1.includes(t1)))
	print('includes: {}'.format(t1.includes(t2)))
	print('includes: {}'.format(t2.includes(t1)))
	print('included: {}'.format(t1.included(t1)))
	print('included: {}'.format(t1.included(t2)))
	print('included: {}'.format(t2.included(t1)))

	print('\n--')
	t = time(t1)  # t1 は timeperiod が設定されている
	print('t1: {0} - {1}'.format(t1.datetime, t1.end.datetime))
	print('t:  {0} - {1}'.format(t.datetime, t.end.datetime))
	t = time(t2)  # t2 は timeperiod が設定されていない
	print('t2: {0} - {1}'.format(t2.datetime, t2.end.datetime))
	print('t:  {0} - {1}'.format(t.datetime, t.end.datetime))

	print('\n--')
	t = time((t1.datetime, 3.0))  # timeperiod を指定して設定．tuple 形式で初期化指定
	dprint('t.datetime:     {}'.format(t.datetime))
	dprint('t.timeperiod:   {}'.format(t.timeperiod))
	dprint('t.end.datetime: {}'.format(t.end.datetime))
	print('t:  {0} - {1}'.format(t.datetime, t.end.datetime))
	t = time([t1.datetime, 3.0])  # timeperiod を指定して設定．list 形式で初期化指定
	dprint('t.datetime:     {}'.format(t.datetime))
	dprint('t.timeperiod:   {}'.format(t.timeperiod))
	dprint('t.end.datetime: {}'.format(t.end.datetime))
	print('t:  {0} - {1}'.format(t.datetime, t.end.datetime))
	t = time(t1, 5.0)  # timeperiod を指定して設定．time 形式と tp=float で初期化指定
	dprint('t.datetime:     {}'.format(t.datetime))
	dprint('t.timeperiod:   {}'.format(t.timeperiod))
	dprint('t.end.datetime: {}'.format(t.end.datetime))
	print('t:  {0} - {1}'.format(t.datetime, t.end.datetime))
	t = time((t1.datetime, 3.0), 5.0)  # timeperiod を指定して設定．tuple 形式と tp=float でで初期化指定
	dprint('t.datetime:     {}'.format(t.datetime))
	dprint('t.timeperiod:   {}'.format(t.timeperiod))
	dprint('t.end.datetime: {}'.format(t.end.datetime))
	print('t:  {0} - {1}'.format(t.datetime, t.end.datetime))
	t = time([t1.datetime, 3.0], 5.0)  # timeperiod を指定して設定．list 形式と tp=float でで初期化指定
	dprint('t.datetime:     {}'.format(t.datetime))
	dprint('t.timeperiod:   {}'.format(t.timeperiod))
	dprint('t.end.datetime: {}'.format(t.end.datetime))
	print('t:  {0} - {1}'.format(t.datetime, t.end.datetime))



#
# [name] nkj.stat.py
# [reference] https://qiita.com/suaaa7/items/745ac1ca0a8d6753cf60
#
# Written by Yoshikazu NAKAJIMA (Wed Apr 14 18:59:31 JST 2021)
#
_LIB_DEBUGLEVEL = 1

import numpy as np
from scipy import stats

import os
import sys
sys.path.append(os.path.abspath(".."))
from nkj.str import *

def ftest(A, B):
	A_var = np.var(A, ddof=1) # A の不偏分散
	B_var = np.var(B, ddof=1) # B の不偏分散
	A_df = len(A) - 1 # A の自由度
	B_df = len(B) - 1 # A の自由度
	f = A_var / B_var # F 比の値
	onesided_pval1 = stats.f.cdf(f, A_df, B_df) # 片側検定の p 値 1
	onesided_pval2 = stats.f.sf(f, A_df, B_df)  # 片側検定の p 値 2
	twosided_pval = min(onesided_pval1, onesided_pval2) * 2 # 両側検定の p 値
	ldprint3(["F:       {0}".format(round(f, 3))])
	ldprint3(["p-value: {0}".format(round(twosided_pval, 4))])
	return f, twosided_pval

def ftest_msg(A, B):
	f, pval = ftest(A, B)
	ldprint(["F:       {0}".format(round(f, 3))])
	ldprint(["p-value: {0}".format(round(pval, 4))])
	print("F 検定帰無仮説：「2 群は等分散である．」")
	if (pval > _PVAL):
		print("p > {0} より、帰無仮説は棄却されませんでした．2 群間は少なくとも不等分散でないと言って良い(= 等分散と言える可能せが残る)．".format(_PVAL))
		print("T-test で検定してください．")
	else:
		print("p <= {0} より、帰無仮説は棄却されました．2 群間は不等分散と言って良い．".format(_PVAL))
		print("Welch t-test で検定してください．")

def ttest(A, B, rel=False):
	if (rel):
		return stats.ttest_rel(A, B) # 対応あり t-test
	else:
		return stats.ttest_ind(A, B) # 対応なし t-test

def welchttest(A, B):
	return stats.ttest_ind(A, B, equal_var=False) # equal_var=False を指定する．

def wilcoxon(A, B, rel=True):
	if (rel):
		return stats.wilcoxon(A, B)     # Wilcoxon 符号付順位和検定．(対応あり)
	else:
		return stats.mannwhitneyu(A, B, alternative='two-sided') # Mann-Whitney u-test, Wilcoxon 順位和検定．(独立=対応なし)

def mannwhitneyu(A, B):
	return wilcoxon(A, B, False)

def pvalcheck(pval, flag=True):
	stars = 3
	for val in [0.001, 0.01, 0.05]:
		if (pval < val):
			if (flag):
				print("< {0}".format(val))
			break
		else:
			stars -= 1
	return stars

def starstr(pval, flag=True):
	if (flag):
		stars = pvalcheck(pval, False)
	else:
		stars = pval
	str = ''
	for i in range(stars):
		str += '*'
	return str

#-- main

_PVAL = 0.05

if __name__ == '__main__':
	lib_debuglevel(_LIB_DEBUGLEVEL)

	#
	# §．2 群間の有意差検定
	#
	# 正規分布が仮定できるとき，
	# 2 群が等分散のとき、t-test (対応あり/なし）
	# 2 群が不等分散のとき、welch t-test (対応あり/なし）
	#
	# 正規分布が仮定できないとき(=ノンパラメトリック分布のとき）、
	# 2 群間に対応ありのとき、Wilcoxon signed-rank test (符号付き順位検定)
	# 2 群間に対応なしのとき、Mann-Whitney U test (Wilcoxon rank sum test(順位和検定))
	#
	# §．3 群以上の間の有意差検定
	#
	# ANalysis Of VAriance (ANOVA)(分散分析): 全体的な平均値の相違を調べる方法で、どの群間に有意差があるかは把握できない．
	#
	print("-- t-test")
	A = np.array([6.3, 8.1, 9.4, 10.4, 8.6, 10.5, 10.2, 10.5, 10.0, 8.8])
	B = np.array([4.8, 2.1, 5.1, 2.0, 4.0, 1.0, 3.4, 2.7, 5.1, 1.4, 1.6])
	ftest_msg(A, B)
	test_result = ttest(A, B) # 対応なし t-test
	print("T-test:    {0}".format(test_result))
	print("statistic: {0}".format(test_result.statistic))
	print("p-value:   {0}".format(test_result.pvalue))
	stars = pvalcheck(test_result.pvalue)
	#print("stars:     {0}".format(stars))
	print("stars:     {0}".format(starstr(test_result.pvalue)))
	print("")

	print("-- Welch t-test")
	A = np.array([13.8, 10.2, 4.6, 10.0, 4.2, 16.1, 14.4, 4.9, 7.7, 11.4])
	B = np.array([3.3, 2.6, 4.0, 4.7, 1.9, 2.9, 4.7, 5.3, 4.3, 3.0, 2.0])
	ftest_msg(A, B)
	test_result = welchttest(A, B)
	print("T-test:    {0}".format(test_result))
	print("statistic: {0}".format(test_result.statistic))
	print("p-value:   {0}".format(test_result.pvalue))
	stars = pvalcheck(test_result.pvalue)
	#print("stars:     {0}".format(stars))
	print("stars:     {0}".format(starstr(test_result.pvalue)))
	print("")

	print("-- Wilcoxon test")
	A = np.array([1.83, 1.50, 1.62, 2.48, 1.68, 1.88, 1.55, 3.06, 1.30])
	B = np.array([0.88, 0.65, 0.60, 1.05, 1.06, 1.29, 1.06, 2.14, 1.29])
	test_result = wilcoxon(A, B)
	print("T-test:    {0}".format(test_result))
	print("statistic: {0}".format(test_result.statistic))
	print("p-value:   {0}".format(test_result.pvalue))
	stars = pvalcheck(test_result.pvalue)
	#print("stars:     {0}".format(stars))
	print("stars:     {0}".format(starstr(test_result.pvalue)))
	print("")

	print("-- Mann-Whitney U test")
	A = np.array([1.83, 1.50, 1.62, 2.48, 1.68, 1.88, 1.55, 3.06, 1.30, 2.01, 3.11])
	B = np.array([0.88, 0.65, 0.60, 1.05, 1.06, 1.29, 1.06, 2.14, 1.29])
	test_result = mannwhitneyu(A, B)
	print("T-test:    {0}".format(test_result))
	print("statistic: {0}".format(test_result.statistic))
	print("p-value:   {0}".format(test_result.pvalue))
	stars = pvalcheck(test_result.pvalue)
	#print("stars:     {0}".format(stars))
	print("stars:     {0}".format(starstr(test_result.pvalue)))
	print("")
